#include <stdlib.h>
#include "game.h"
//#include "mem.h"
#include <io.h>
#include <fcntl.h>
#include <sys\stat.h>
#include <assert.h>
#include <stdio.h>
#include <ctype.h>

#define LEVEL_CHECKSUM 2816
#define ALTLEVEL_CHECKSUM 2959

//#define SHOW_CHECKSUM
//#define easyCheats

const int BLANK = 0;
const int PLATFORM = 1;
const int LADDER = 2;
const int EGG = 3;
const int GRAIN = 4;
const int INV_PLATFORM = 5;
const int SOLID_LADDER = 6;

/* save file format: (*.chuckie)
	0   CHUCKxxx/n - id header

	levels file string length
	levels file string
	padding

	8   level checksum
	12  no. of players
	16  current player
	20  state
	24  statevar
	28  elevatorPos
	32  paused
	36  halt bonus
	40  time
	44  runner (14*4 = 56)
	100 chuckie (28)

	128 Map*noOfPlayers (2160+4+128+8 = 2300)

	(for each player)
	0 score
	4 lives
	8 level
	12 bonus  (16)

	states acceptable for save : PLAYING, DONELEVEL, STARTINGGAME, PLAYERDEATH
								GAMEOVER, PLAYERPROMPT

	%% Reasons for not giving score code %%

	- Level jump (.ini) used
	- Alternative lives (.ini) used
	- Levels.txt changed
	- Level jump cheat
	- Cage Chuckie cheat used
	- Invulnerability cheat used
	- Frame rate too low
	- Game load used
	- extended jumps used

*/

#define cp currentPlayer

Game::Game()
{
	spSwans=NULL;
	spBlocks=NULL;
	spRunner=NULL;
	spLevel=NULL;
	spNumbers=NULL;
	spFrontend=NULL;
	int c;
	for(c=0; c<5; c++)
		spFont[c]=NULL;
	soEgg=NULL;
	soGrain=NULL;
	soWalk=NULL;
	soJump=NULL;
	soDeath=NULL;
	soDoneLevel=NULL;
	soLadder=NULL;
	soFall=NULL;
}

Game::~Game()
{
	/*
	if (spSwans!=NULL) delete spSwans;
	if (spBlocks!=NULL) delete spBlocks;
	if (spRunner!=NULL) delete spRunner;
	if (spLevel!=NULL) delete spLevel;
	int c;
	for(c=0; c<5; c++)
	{
		if (spFont[c]!=NULL) delete spFont[c];
	}
	if (spNumbers!=NULL) delete spNumbers;
	if (spFrontend!=NULL) delete spFrontend;
	for(c=0; c<5; c++)
		if (spFont[c]!=NULL) delete spFont[c];

	if (maps!=NULL) delete[] maps;

	if (soEgg!=NULL) delete soEgg;
	if (soGrain!=NULL) delete soGrain;
	if (soWalk!=NULL) delete soWalk;
	if (soJump!=NULL) delete soJump;
	if (soFall!=NULL) delete soFall;
	if (soDeath!=NULL) delete soDeath;
	if (soLadder!=NULL) delete soLadder;
	if (soDoneLevel!=NULL) delete soDoneLevel;
	*/
}

BOOL Game::init(BOOL fs, int xs, int ys)
{
	state=DRAWINGMENU; stateVar=0;
	invulnerable=FALSE;
	levJump=FALSE;
	showFps=0;
	shiftPressed=FALSE;
	options.cheating=FALSE;
	options.topScore=0;
	options.topLevel=0;

	if (fs)
	{
		magnify=2;
		xmod=30;
		ymod=20;
	}
	else
	{
		magnify=xs/320;
		xmod=0;
		ymod=0;
	}
	fullScreen=fs;

	int c;
	for(c=0; c<4; c++)
	{
		score[c]=0;
	}

	FileReader* ini = new FileReader("chuckie.ini");
	if (!ini->valid())
	{
		showMessage("Fatal error: Cannot load chuckie.ini");
		return FALSE;
	}
	ini->setSection("[Chuckie]");

	moveLeftKey=(char)(ini->readInt("LeftKey"));
	moveRightKey=(char)(ini->readInt("RightKey"));
	moveUpKey=(char)(ini->readInt("UpKey"));
	moveDownKey=(char)(ini->readInt("DownKey"));
	jumpKey=(char)(ini->readInt("JumpKey"));
	pauseKey=(char)(ini->readInt("PauseKey"));
	quitKey=(char)(ini->readInt("QuitKey"));

	options.framesPerSecond=ini->readInt("FramesPerSecond");
	timePerFrame=1000/options.framesPerSecond;
	options.useJoystick=ini->readInt("UseJoystick");
	options.levelJump=ini->readInt("LevelJump")-1;
	options.startLives=ini->readInt("StartLives");
	options.persistentKeys=ini->readInt("PersistentKeys");
	options.persistentScores=ini->readInt("PersistentScores");
	options.extendedJumps=ini->readInt("ExtendedJumps");
	options.soundMode = ini->readInt("SoundMode");
	Sound::setSoundMode(options.soundMode);

	char hs[]="Highscore.";
	char hn[]="Highname.";
	for(c=0; c<10; c++)
	{
		hs[9]=(char)(c+48);
		highScores[c]=ini->readInt(hs);
		hn[8]=(char)(c+48);
		memcpy(highNames[c],ini->readString(hn)+1,8);
		highNames[c][8]='\0';
	}

	if (!loadLevels(ini->readString("LevelFile")))
		loadLevels("levels.txt");

	delete ini;

	leftPressed=FALSE; rightPressed=FALSE; upPressed=FALSE; downPressed=FALSE;
	jumpPressed=FALSE;

	noOfPlayers=1;
	currentPlayer=0;
	paused=FALSE;
	elevatorPos=0;
	for(c=0; c<4; c++)
	{
		level[c]=options.levelJump;
		lives[c]=options.startLives;
		livesDirty[c]=TRUE;
		scoreDirty[c]=TRUE;
	}

    spChuckie = new Sprite;
	spSwans = new Sprite;
	spBlocks = new Sprite;
	spRunner = new Sprite;
	spLevel = new Sprite;
	spNumbers = new Sprite;
	spFrontend = new Sprite;
	for(c=0; c<5; c++)
		spFont[c] = new Sprite;
	loadSprites(magnify);

	soEgg = new Sound;
	soEgg->load("EggSound");
	soGrain = new Sound;
	soGrain->load("GrainSound");
	soDeath = new Sound;
	soDeath->load("DeathSound");
	soLadder = new Sound;
	soLadder->load("LadderSound");
	soDoneLevel = new Sound;
	soDoneLevel->load("DoneLevelSound");
	soJump = new Sound;
	soJump->load("JumpSound");
	soWalk = new Sound;
	soWalk->load("WalkSound");
	soFall = new Sound;
	soFall->load("FallSound");

	nextFrameTime=timeGetTime();

	return TRUE;
}

BOOL Game::loadLevels(char* fname)
{
	FileReader *fr = new FileReader(fname);
	if (!fr->valid())
	{
		showMessage("Fatal error: Cannot load levels data");
		return FALSE;
	}

	fr->setSection("[levels]");

	noOfMaps = fr->readInt("NumberPresent");
	maps = new Map[noOfMaps];

	for(int cc=0; cc<noOfMaps; cc++)
	{
		char l[]="[ ]\0";
		l[1]=(char)cc+49;
		fr->setSection(l);
		for(int y=0; y<27; y++)
			for(int x=0; x<20; x++)
				switch( fr->readString(y)[x] )
				{
				case '0' : maps[cc].matrix[x][y] = BLANK; break;
				case '1' : maps[cc].matrix[x][y] = PLATFORM; break;
				case '2' : maps[cc].matrix[x][y] = LADDER; break;
				case '3' : maps[cc].matrix[x][y] = EGG; break;
				case '4' : maps[cc].matrix[x][y] = GRAIN; break;
				case '5' : maps[cc].matrix[x][y] = INV_PLATFORM; break;
				case '6' : maps[cc].matrix[x][y] = SOLID_LADDER; break;
				}
		maps[cc].elevator = fr->readInt("Elevator");
		maps[cc].runner.x = fr->readInt("StartX");
		maps[cc].runner.y = fr->readInt("StartY");
		char sw[]="Swan ";
		for(int swan=0; swan<8; swan++)
		{
			sw[4]=(char)(swan+49);
			char* str=fr->readString(sw);
			if (strlen(str)==1)
			{
				maps[cc].swans[swan].x=-1;
				maps[cc].swans[swan].y=-1;
			}
			else
			{
				int ptr=0;
				if (!isdigit(str[0]))
				{ showMessage("Cannot read swan data."); return FALSE; }

				if (isdigit(str[1]))
				{
					maps[cc].swans[swan].x=(str[0]-48)*10+str[1]-48;
					ptr=2;
				}
				else
				{
					maps[cc].swans[swan].x=str[0]-48;
					ptr=1;
				}
				if (str[ptr++]!=',')
				{ showMessage("Cannot read swan data."); return FALSE; }

				if (isdigit(str[ptr+1]))
				{
					maps[cc].swans[swan].y=(str[ptr]-48)*10+str[ptr+1]-48;
					ptr+=2;
				}
				else
				{
					maps[cc].swans[swan].y=str[ptr]-48;
					ptr++;
				}

				if (str[ptr]==',')
					maps[cc].swans[swan].dir=str[ptr+1]-48;
				else
					maps[cc].swans[swan].dir=1;

				maps[cc].swans[swan].pecking=0;
			}
		}
	}

	options.check=0;
	for(int cl=0; cl<8; cl++)
	{
		for(int cx=0; cx<20; cx++)
			for(int cy=0; cy<27; cy++)
				options.check+=maps[cl].matrix[cx][cy];
		for(int cs=0; cs<8; cs++)
			options.check = options.check + maps[cl].swans[cs].x + maps[cl].swans[cs].y;
	}

	#ifdef SHOW_CHECKSUM
	showMessage("Level checksum = ",options.check);
	#endif

	delete fr;

	writeProfileStr("LevelFile",fname);

	return TRUE;
}

void Game::loadSprites(int m)
{
	spLevel->setSize(320*m,224*m);

	spFrontend->load("frontend",320*m,163*m);
	spFrontend->clearMask();

	spChuckie->load("chuckie",32*m,96*m);

	spSwans->load("swans",28*m,200*m);

	spRunner->load("runner",16*m,150*m);

	spBlocks->load("blocks",16*m,32*m);
	spBlocks->clearMask();

	spNumbers->load("numbers",100*m,66*m);
	spNumbers->clearMask();

	RECT rect;
	rect.left=0; rect.top=64*m; rect.right=320*m; rect.bottom=104*m;
	for(int c=0; c<5; c++)
	{
		spFont[c]->setSize(320*m,40*m);
		spFrontend->draw(spFont[c],0,0,rect);
		spFont[c]->clearMask();
	}
	spFont[BLUE]->colourize(Sprite::BLUE, spNumbers, 48*m, 18*m, m);
	spFont[CYAN]->colourize(Sprite::CYAN, spNumbers, 48*m, 18*m, m);
	spFont[MAGENTA]->colourize(Sprite::MAGENTA, spNumbers, 48*m, 18*m, m);
	spFont[YELLOW]->colourize(Sprite::YELLOW, spNumbers, 48*m, 18*m, m);
	spFont[GREEN]->colourize(Sprite::GREEN, spNumbers, 48*m, 18*m, m);

}

int Game::getymod(int vel)
{
	switch(vel)
	{
	case -19: return -1;
	case -18: return -4;
	case -17: return -4;
	case -16: return -4;
	case -15: return -4;
	case -14: return -3;
	case -13: return -3;
	case -12: return -3;
	case -11: return -3;
	case -10: return -2;
	case -9: return -2;
	case -8: return -2;
	case -7: if (runner.minijump) return -1; else return -2;
	case -6: return -1;
	case -5: return -1;
	case -4: return -1;
	case -3: return -1;
	case -2: return 0;
	case -1: return 0;
	case 0: return 0;
	case 1: return 0;
	case 2: return 1;
	case 3: return 1;
	case 4: return 1;
	case 5: return 1;
	case 6: return 2;
	case 7: return 2;
	case 8: return 2;
	case 9: return 2;
	default: return (vel>=18)?5:((vel-10)/4)+3;
	}
}

BOOL Game::iterate()
{

static long currentTime;
if (nextFrameTime > (currentTime=timeGetTime())) return FALSE;
while(nextFrameTime<=currentTime)
	nextFrameTime+=timePerFrame;

BOOL fire4Pressed;

// poll joystick here so fire button can start game
upPressed=FALSE; downPressed=FALSE; leftPressed=FALSE; rightPressed=FALSE; jumpPressed=FALSE;
fire1Pressed=FALSE; fire2Pressed=FALSE;
fire3Pressed=FALSE; fire4Pressed=FALSE;

if (options.useJoystick)
{

	static JOYINFOEX joyinfo;
	memset(&joyinfo,'/0',sizeof(joyinfo));
	joyinfo.dwFlags=JOY_RETURNBUTTONS|JOY_RETURNX|JOY_RETURNY;
	joyinfo.dwSize=sizeof(joyinfo);
	static MMRESULT mmresult;
	mmresult = joyGetPosEx(JOYSTICKID1,&joyinfo);

	if (mmresult==JOYERR_NOERROR)
	{
		if (joyinfo.dwButtons!=0 && joyinfo.dwButtons!=8) {jumpPressed=TRUE;fire1Pressed=TRUE;}
		if ((joyinfo.dwButtons&2)==2) fire2Pressed=TRUE;
		if ((joyinfo.dwButtons&4)==4) fire3Pressed=TRUE;
		if ((joyinfo.dwButtons&8)==8) fire4Pressed=TRUE;
		if (joyinfo.dwXpos<20000) leftPressed=TRUE;
		if (joyinfo.dwXpos>40000) rightPressed=TRUE;
		if (joyinfo.dwYpos<20000) upPressed=TRUE;
		if (joyinfo.dwYpos>40000) downPressed=TRUE;
	}
}

switch(state)
{
case PLAYING:
	if (paused) return TRUE;

	if (GetAsyncKeyState(moveUpKey)<0) upPressed=TRUE;
	if (GetAsyncKeyState(moveDownKey)<0) downPressed=TRUE;
	if (GetAsyncKeyState(moveLeftKey)<0) leftPressed=TRUE;
	if (GetAsyncKeyState(moveRightKey)<0) rightPressed=TRUE;
	if (GetAsyncKeyState(jumpKey)<0) jumpPressed=TRUE;

	static int timeCount;
	if (timeCount++%8==0)
		if (haltBonus>0)
			haltBonus--;
		else
		{
			time--;
			timeDirty=TRUE;
			if (time<0) {time=0;die();}
			if (time%5==0)
			{
				bonus[currentPlayer]-=10;
				bonusDirty=TRUE;
				if (bonus[currentPlayer]<0) bonus[currentPlayer]=0;
			}
		}

	static BOOL canWalk;
	static int oldX;
	static int oldY;

	oldX=runner.x;
	oldY=runner.y;

	canWalk=FALSE;
	if ((!runner.onLadder && runner.y%8==0) || runner.onElevator) canWalk=TRUE;
	if (runner.onLadder && runner.y%8==0 &&
			( matrix((runner.x/8)+1,(runner.y/8)+1)==PLATFORM
			 || matrix((runner.x/8)-1,(runner.y/8)+1)==PLATFORM))
		canWalk=TRUE;

	if (runner.jumping) canWalk=FALSE;

	static int walking;
	walking=FALSE;

	// walk right
	if (rightPressed && canWalk && matrix(((runner.x-2)/8)+1,runner.y/8)!=PLATFORM)
	{
		runner.onLadder=FALSE;
		runner.x+=1;
		runner.dir=1;
		runner.jumpFromLadder=FALSE;
		if (!soWalk->playing() && !jumpPressed) soWalk->playLooped();
		if (runner.x>152)
			runner.x=152;
		else
			walking=TRUE;
	}

	// walk left
	if (leftPressed && canWalk && matrix(((runner.x+1)/8),runner.y/8)!=PLATFORM)
	{
		runner.dir=0;
		runner.onLadder=FALSE;
		runner.x-=1;
		runner.jumpFromLadder=FALSE;
		if (!soWalk->playing() && !jumpPressed) soWalk->playLooped();
		if (runner.x<0)
			runner.x=0;
		else
			walking=TRUE;
	}

	if (!walking) soWalk->stop();
	if (canWalk) soFall->stop();

	// jump left/right
	if ( jumpPressed && (leftPressed||rightPressed) && ( canWalk || runner.onLadder) )
	{
		soJump->stop();soJump->play();
		runner.dir=leftPressed?0:1;
		runner.jumping=TRUE;
		runner.yvel=-10;
		if (options.extendedJumps && (upPressed||fire2Pressed) && level[currentPlayer]>=noOfMaps) runner.yvel=-12;
		if (options.extendedJumps && ((upPressed&&downPressed)||fire3Pressed) && level[currentPlayer]>=noOfMaps*2) runner.yvel=-19;
		runner.xvel=leftPressed?-1:1;
		runner.jumpFromLadder=FALSE;
		runner.fallNudge=0;
		if (matrix(((runner.x+(leftPressed?1:-2))/8+(leftPressed?0:1)),runner.y/8)==PLATFORM)
			runner.minijump=TRUE;

		if (leftPressed && isPlatform((runner.x)/8-2,runner.y/8) && !runner.onLadder
			&& !isPlatform((runner.x)/8-1,runner.y/8)) {runner.byebye=14;soDeath->play();}
		if (rightPressed && matrix((runner.x-2)/8+3,runner.y/8)==PLATFORM
			&& matrix((runner.x-2)/8+2,runner.y/8)!=PLATFORM
			&& matrix((runner.x-2)/8,runner.y/8+1)==PLATFORM ) runner.byebye=13;

		runner.onLadder=FALSE;
	}

	// jump straight
	if ( jumpPressed && !leftPressed && !rightPressed )
	{
		if (!runner.jumping) {soJump->stop();soJump->play();}
		if (canWalk && !runner.onLadder)
		{
			runner.yvel=-10;
			if (options.extendedJumps && (upPressed||fire2Pressed) && level[currentPlayer]>=noOfMaps) runner.yvel=-12;
			if (options.extendedJumps && ((upPressed&&downPressed)||fire3Pressed) && level[currentPlayer]>=noOfMaps*2) runner.yvel=-19;
			runner.xvel=0;
			runner.onLadder=FALSE;
			runner.jumping=TRUE;
			runner.fallNudge=0;
			if (matrix(runner.x/8,runner.y/8)==PLATFORM)
				runner.minijump=TRUE;
		}
		if (runner.onLadder)// && !upPressed && !downPressed)
		{
			runner.yvel=-10;
			if (options.extendedJumps && (upPressed||fire2Pressed) && level[currentPlayer]>=noOfMaps) runner.yvel=-12;
			if (options.extendedJumps && ((upPressed&&downPressed)||fire3Pressed) && level[currentPlayer]>=noOfMaps*2) runner.yvel=-19;
			runner.xvel=0;
			runner.jumping=TRUE;
			runner.jumpFromLadder=TRUE;
			runner.onLadder=FALSE;
			runner.fallNudge=0;
		}
	}

	static BOOL ladder;
	ladder=FALSE; // used for sound effect

	// normal climb ladder
	if (upPressed && runner.x%8==0 && (
			isLadder(runner.x/8, runner.y/8-(((runner.y%8)==0)?2:1) )
			|| (runner.jumping && isLadder(runner.x/8,runner.y/8-1))
			|| (runner.jumping && isLadder(runner.x/8,runner.y/8))
		))
	{
//		if (runner.jumping && isPlatform(runner.x/8,runner.y/8-1)) goto skip;
		if (runner.jumping && jumpPressed && (leftPressed || rightPressed))
		{
			soJump->stop(); soJump->play();
			runner.yvel=-10;
			runner.xvel=leftPressed?-1:1;
			runner.dir=leftPressed?0:1;
            runner.jumpFromLadder=FALSE;
		}
		else
		{
			if (runner.jumping && jumpPressed && !leftPressed
				&& !rightPressed && runner.xvel!=0)
				runner.squirm=runner.dir;

			runner.y-=2;
			if (options.extendedJumps && level[currentPlayer]>=noOfMaps && jumpPressed && runner.y%4==0) runner.y-=2;
			if (runner.jumping && runner.y%2==1) runner.y--;
			runner.onLadder=TRUE;
			runner.jumping=FALSE;
			runner.anim++;runner.anim%=4;
			runner.jumpFromLadder=FALSE;
			if (!soLadder->playing()) soLadder->playLooped();
			ladder=TRUE;
		}
	}
//	skip:

	// normal descend ladder
	if (downPressed && runner.x%8==0 //&& matrix(runner.x/8,runner.y/8)==LADDER)
		&& isLadder(runner.x/8,runner.y/8+1))
	{
		if (runner.jumping && upPressed && jumpPressed && (leftPressed || rightPressed))
		{
			soJump->play();
			runner.yvel=-10;
			runner.xvel=leftPressed?-1:1;
			runner.dir=leftPressed?0:1;
		}
		else
		{
			runner.y+=2;
			if (runner.jumping && runner.y%2==1) runner.y--;
			runner.onLadder=TRUE;
			runner.jumping=FALSE;
			runner.anim++; runner.anim%=4;
			runner.jumpFromLadder=FALSE;
			if (!soLadder->playing()) soLadder->playLooped();
			ladder=TRUE;
		}
	}

	if (!ladder) soLadder->stop(); else soJump->stop();
	if (runner.jumpFromLadder) runner.anim=0;

	// continue existing jump/fall
	if (runner.jumping==TRUE)
	{
		runner.y+=getymod(runner.yvel);
		runner.x+=runner.xvel;
		runner.yvel++;
		if (runner.x<0) { runner.x=1; runner.xvel=1; }
		if (runner.x>152) { runner.x=151; runner.xvel=-1; }
		if (runner.minijump && runner.yvel==-5)
		{
			runner.jumping=FALSE;
			runner.minijump=FALSE;
		}
		if (runner.fallNudge>0) {runner.x++;runner.fallNudge--;}
		if (runner.fallNudge<0) {runner.x--;runner.fallNudge++;}
	}

	// running along the ceiling (extended jumps)
	if (level[currentPlayer]>=noOfMaps*2 && runner.y<6 && runner.yvel==-5 && jumpPressed
		&& ((upPressed&&downPressed)||fire3Pressed) )
	{
		runner.yvel-=2;
		runner.y+=2;
	}
	if (runner.y<=2) runner.yvel=2;

	// bounce off walls when jumping
	if (runner.jumping && runner.xvel==1 && runner.x%8==1 && runner.yvel>-2)
		if (matrix(runner.x/8+1,runner.y/8)==PLATFORM
			|| (runner.yvel>=8 && matrix(runner.x/8+1,runner.y/8+1)==PLATFORM)
			|| (runner.yvel>=8 && matrix(runner.x/8+1,runner.y/8-1)==PLATFORM))
			runner.xvel=-1;

	if (runner.jumping && runner.xvel==-1 && runner.x%8==7 && runner.yvel>-2)
		if (matrix(runner.x/8,runner.y/8)==PLATFORM
			|| (runner.yvel>=8 && matrix(runner.x/8,runner.y/8-1)==PLATFORM)
			|| (runner.yvel>=8 && matrix(runner.x/8,runner.y/8+1)==PLATFORM))
			runner.xvel=1;

	if (runner.byebye>0)
	{
		runner.byebye--;
		if (runner.byebye==0) runner.xvel=0-runner.xvel;
	}

	// falling off the edge
	if (!runner.onLadder && !runner.jumping && !jumpPressed && !runner.onElevator)
	{
		if (runner.x%8>3 && (matrix(runner.x/8+1,runner.y/8+1)==BLANK
			|| matrix(runner.x/8+1,runner.y/8+1)==EGG
			|| matrix(runner.x/8+1,runner.y/8+1)==GRAIN))
		{
			runner.yvel=8;
			runner.jumping=TRUE;
			runner.xvel=0;
			runner.fallNudge=(runner.dir*4)-2;
			soFall->play();
		}

		if (runner.x%8<5 && (matrix(runner.x/8,runner.y/8+1)==BLANK
			|| matrix(runner.x/8,runner.y/8+1)==EGG
			|| matrix(runner.x/8,runner.y/8+1)==GRAIN))
		{
			runner.yvel=8;
			runner.jumping=TRUE;
			runner.xvel=0;
			runner.fallNudge=(runner.dir*4)-2;
			soFall->play();
		}
	}

	// check for landings
	if (runner.jumping && runner.yvel>=0)
	{
		static int tx,ty;
		tx=runner.x/8;
		ty=(runner.y+getymod(runner.yvel)+1)/8+1;
		if (runner.x%8>3 && isPlatform(tx+1,ty) )
		{
			runner.yvel=0;
			runner.jumping=FALSE;
			if (runner.y%8!=0) runner.y = ((runner.y/8)+1)*8;
			if (rightPressed) runner.x++;
			if (leftPressed) runner.x--;
		}
		if (runner.x%8<5 && isPlatform(tx,ty) )
		{
			runner.yvel=0;
			runner.jumping=FALSE;
			if (runner.y%8!=0) runner.y = ((runner.y/8)+1)*8;
			if (rightPressed) runner.x++;
			if (leftPressed) runner.x--;
		}
	}

	if (!jumpPressed || (runner.jumping && runner.xvel!=0)) runner.squirm=-1;

	// land on elavator
	static int elevator2Pos;
	runner.onElevator=FALSE;
	if (map[cp].elevator!=0) // && runner.jumping)
	{
		elevator2Pos=elevatorPos-80;
		if (elevator2Pos<0) elevator2Pos+=226;
		if (runner.x > map[cp].elevator*8-8 && runner.x < map[cp].elevator*8)
		{
			static int elevatorHeight; // so a plummeting runner won't pass thru elevator
			if (getymod(runner.yvel) > 4) elevatorHeight=getymod(runner.yvel);
			else elevatorHeight=4;
			if (elevatorPos-16<=runner.y && elevatorPos+elevatorHeight-16>=runner.y)
			{
				runner.y=elevatorPos-16;
				runner.jumping=FALSE;
				runner.onElevator=TRUE;
			}
			if (elevator2Pos-16<=runner.y && elevator2Pos+elevatorHeight-16>=runner.y)
			{
				runner.y=elevator2Pos-16;
				runner.jumping=FALSE;
				runner.onElevator=TRUE;
			}
		}
	}

	// stop jump sound if we have landed
	if (!runner.jumping) {soJump->stop();soFall->stop();}

	// test for egg/grain
	if (runner.x%8>4)
	{
		pickUp(runner.x/8+1,runner.y/8);
		if (runner.y%8>0) pickUp(runner.x/8+1,runner.y/8-1);
	}
	if (runner.x%8<4)
	{
		pickUp(runner.x/8,runner.y/8);
		if (runner.y%8>0) pickUp(runner.x/8,runner.y/8-1);
	}

	if ((!runner.onLadder && oldX!=runner.x) )//|| (runner.onLadder && oldY!=runner.y))
	{
		runner.anim++;
		runner.anim%=4;
	}

	// swans movement
	static int swanCount,tx,ty,td,c,exitCount,swanTemp,pecking;
	static int first,second,third;
	static BOOL exits[4];
	if (swanCount++%10==0)
		for(c=0; c<8; c++)
			if (map[cp].swans[c].x!=-1)
			{
				tx=map[cp].swans[c].x; ty=map[cp].swans[c].y;
				td=map[cp].swans[c].dir; pecking=map[cp].swans[c].pecking;
				if (pecking>0)
				{
					pecking++;
					if (pecking==5) pecking=0;
					if (pecking==2)
					{
						setMatrix(tx/8+(td==1?1:-1),ty/8,BLANK);
						buildLevelSprite(currentPlayer);
					}
				}
				else if (tx%8==0 && ty%8==0)
				{
					memset(exits,'\0',sizeof(BOOL)*4); exitCount=0;
					if (matrix(tx/8,ty/8-2)==LADDER) {exits[0]=TRUE;exitCount++;}
					if (matrix(tx/8,ty/8+1)==LADDER) {exits[2]=TRUE;exitCount++;}
					if (matrix(tx/8+1,ty/8+1)==PLATFORM || (matrix(tx/8+1,ty/8+1)==LADDER
						&& ( matrix(tx/8,ty/8+1)==PLATFORM || matrix(tx/8+2,ty/8+1)==PLATFORM ) ) )
					{exits[1]=TRUE;exitCount++;}
					if (matrix(tx/8-1,ty/8+1)==PLATFORM || (matrix(tx/8-1,ty/8+1)==LADDER
						&& ( matrix(tx/8,ty/8+1)==PLATFORM || matrix(tx/8-2,ty/8+1)==PLATFORM ) ) )
					{exits[3]=TRUE;exitCount++;}

					// make level 25 easier
					if (options.check==LEVEL_CHECKSUM && exitCount>=3
						&& exits[0]==TRUE && exits[2]==TRUE && level[cp]%noOfMaps==0
						&& level[cp]>=noOfMaps*3)
						if ((rand()%3)==0) exits[0]=FALSE;

					switch(exitCount)
					{
					case 1: td=td^2; break;
					case 2:
						if (!exits[td])
						{
							exits[td^2]=FALSE;
							for(swanTemp=0; swanTemp<4; swanTemp++)
								if (exits[swanTemp]) td=swanTemp;
						}
						break;
					case 3:
						exits[td^2]=FALSE;
						first=-1;
						for(swanTemp=0; swanTemp<4; swanTemp++)
						{
							if (exits[swanTemp] && first==-1) first=swanTemp;
							if (exits[swanTemp] && first>-1) second=swanTemp;
						}
						if ((rand()%2)==0) td=first; else td=second;
						break;
					case 4:
						exits[td^2]=FALSE;
						first=-1;second=-1;
						for(swanTemp=0; swanTemp<4; swanTemp++)
						{
							if (exits[swanTemp] && first==-1 && second==-1) first=swanTemp;
							if (exits[swanTemp] && first>-1 && second==-1) second=swanTemp;
							if (exits[swanTemp] && first>-1 && second>-1) third=swanTemp;
						}
						if ((rand()%3)==0)
							td=first;
						else
							td=((rand()%2)==0)?second:third;
						break;
					}
					if ( (td==1 && matrix(tx/8+1,ty/8)==GRAIN)
						|| (td==3 && matrix(tx/8-1,ty/8)==GRAIN) ) pecking=1;
				}
				if (pecking==0)
				{
					tx = tx + ( td==1?4:0 ) - ( td==3?4:0 );
					ty = ty + ( td==2?4:0 ) - ( td==0?4:0 );
				}
				map[cp].swans[c].x=tx; map[cp].swans[c].y=ty;
				map[cp].swans[c].dir=td; map[cp].swans[c].pecking=pecking;
			}

	// death by swan
	if (!invulnerable && state!=DONELEVEL)
		for(c=0; c<8; c++)
			if (map[cp].swans[c].x!=-1)
				if (map[cp].swans[c].x+6>=runner.x && map[cp].swans[c].x<=runner.x+7
					&& map[cp].swans[c].y+11>=runner.y && map[cp].swans[c].y<=runner.y+15)
						die();

	// chuckie
	static int chuckieCount;
	if (!chuckie.caged && chuckieCount++%10==0)
	{
		if (runner.x+4>chuckie.x+8)
		{	chuckie.xvel++; chuckie.dir=0; }
		else
		{	chuckie.xvel-=1; chuckie.dir=1; }
		if (runner.y+8>chuckie.y+12) chuckie.yvel++; else chuckie.yvel--;
		if (chuckie.xvel>5) chuckie.xvel=5;
		if (chuckie.xvel<-5) chuckie.xvel=-5;
		if (chuckie.yvel>5) chuckie.yvel=5;
		if (chuckie.yvel<-5) chuckie.yvel=-5;
		chuckie.x+=chuckie.xvel;
		chuckie.y+=chuckie.yvel;
		if (chuckie.x<0 || chuckie.x>144) {chuckie.xvel=-chuckie.xvel;chuckie.x+=chuckie.xvel*2;}
		if (chuckie.y<0 || chuckie.y>190) {chuckie.yvel=-chuckie.yvel;chuckie.y+=chuckie.yvel*2;}
	}

	// death by chuckie
	if (!chuckie.caged && !invulnerable && state!=DONELEVEL)
		if (chuckie.x+14>runner.x && chuckie.x<runner.x+3
			&& chuckie.y+21>runner.y && chuckie.y<runner.y+9)
		die();

	if (!runner.onLadder && oldX==runner.x)
		runner.anim=0;

	// crushed head or depth plummet
	if (state!=DONELEVEL &&
		(runner.y>213 || (runner.y<6 && runner.onElevator && !jumpPressed))) die();

	chuckie.anim++;
	chuckie.anim%=8;

	elevatorPos--;
	if (elevatorPos<0) elevatorPos=225;
	break;

case MENU:

	// Rolling demo:
	// 0-499 menu
	// 500-749 level
	// 750-1049 keys
	// 1050-1300 level

	if (fire1Pressed||fire2Pressed||fire3Pressed||fire4Pressed) {stateVar=0;state=PLAYERPROMPT;Sprite::cls();}

	stateVar++;
	if (stateVar==500 || stateVar==1050)
	{
		Sprite::cls();
		currentPlayer=0;
		setLevel(0,rand()%8);
		buildLevelSprite(0);
	}
	if (stateVar==750)
		Sprite::cls();
	if (stateVar==1300)
	{
		Sprite::cls();
		state=DRAWINGMENU;
		stateVar=0;
	}
	break;
case DRAWINGMENU:
	if (fire1Pressed||fire2Pressed||fire3Pressed||fire4Pressed) {stateVar=0;state=PLAYERPROMPT;Sprite::cls();}

	static int drawMenuCount;
	if (drawMenuCount++%2==0)
	{
		stateVar++;
		if (stateVar==18) {state=MENU;stateVar=0;}
	}
	break;
case STARTINGGAME:
	stateVar++;
	if (stateVar==90)
	{
		Sprite::cls();
		resetLevel(currentPlayer);
		buildLevelSprite(currentPlayer);
	}
	if (stateVar==110)
	{
		state=PLAYING;
		if (!invulnerable) lives[currentPlayer]--;
		livesDirty[currentPlayer]=TRUE;
	}
	break;
case PLAYERDEATH:
	stateVar++;
	if (stateVar==100)
	{
		if (lives[currentPlayer]==0)
		{
			if (score[currentPlayer]>options.topScore)
			{
				options.topScore=score[currentPlayer];
				options.topLevel=level[currentPlayer];
			}
			state=GAMEOVER;
			stateVar=0;
		}
		else
		{
			scoreDirty[currentPlayer]=TRUE;
			if (noOfPlayers>1)
				do {currentPlayer++;currentPlayer%=noOfPlayers;} while(lives[currentPlayer]==0);
			scoreDirty[currentPlayer]=TRUE;
			resetLevel(currentPlayer);
			Sprite::cls();
			state=STARTINGGAME;
			stateVar=0;
		}
	}
	break;
case PLAYERPROMPT:
	if (fire1Pressed||fire2Pressed||fire3Pressed||fire4Pressed)
	{
		stateVar=1;
		noOfPlayers=1;
		level[0]=options.levelJump;
	}

	if (stateVar>0) stateVar++;
	if (stateVar>50)
	{
		Sprite::cls();
		static int cc;
		for(cc=0; cc<noOfPlayers; cc++)
		{
			setLevel(cc,options.levelJump);
			score[cc]=0;
			lives[cc]=options.startLives;
		}
		state=STARTINGGAME;
		stateVar=0;
		options.cheating=FALSE;
		memset(&options.reasons,'\0',sizeof(options.reasons));
		if (invulnerable) {options.cheating=TRUE;options.reasons.invulnerable=TRUE;}
		if (levJump) {options.cheating=TRUE;options.reasons.levelJump=TRUE;}
		if (options.startLives!=5) {options.cheating=TRUE;options.reasons.startLives=TRUE;}
		if (options.levelJump!=0) {options.cheating=TRUE;options.reasons.startLevel=TRUE;}
		if (options.framesPerSecond<40) {options.cheating=TRUE;options.reasons.tooSlow=TRUE;}
		if (options.extendedJumps) {options.cheating=TRUE;options.reasons.extJumps=TRUE;}
		if (options.check!=LEVEL_CHECKSUM && options.check!=ALTLEVEL_CHECKSUM)
		{
			options.cheating=TRUE;
			options.reasons.levelsChanged=TRUE;
		}
	}
	break;
case GAMEOVER:
	stateVar++;
	if (stateVar==149 && score[currentPlayer]>=highScores[9])
		{Sprite::cls();state=HIGHSCORE;stateVar=0;}
	if (stateVar==150)
	{
		{
			Sprite::cls();
			static int playersLeft,cc;
			playersLeft=0;
			for(cc=0;cc<noOfPlayers;cc++) if (lives[cc]>0) playersLeft++;
			if (playersLeft==0)
			{
				state=DRAWINGMENU;
				stateVar=0;
			}
			else
			{
				scoreDirty[currentPlayer]=TRUE;
				livesDirty[currentPlayer]=TRUE;
				while(lives[currentPlayer]==0) {currentPlayer++;currentPlayer%=noOfPlayers;}
				scoreDirty[currentPlayer]=TRUE;
				resetLevel(currentPlayer);
				Sprite::cls();
				state=STARTINGGAME;
				stateVar=0;
			}
		}
	}
	break;
case HIGHSCORE:
	if (stateVar==0) // determine score position
	{
		ranking=9;
		while( score[currentPlayer]>=highScores[ranking-1] && ranking>0) ranking--;
		static int hsc;
		if (ranking<9)
			for(hsc=9; hsc>ranking; hsc--)
			{
				strcpy(highNames[hsc],highNames[hsc-1]);
				highScores[hsc]=highScores[hsc-1];
			}
		highScores[ranking]=score[currentPlayer];
		strcpy(highNames[ranking],"        ");
		stateVar=1;
	}
	break;
case DONELEVEL:
	if (stateVar==0)
	{
		static int bonusMod;
		bonusMod = (level[currentPlayer]<9)?(level[currentPlayer]+1)*10:100;
		bonus[currentPlayer]-=bonusMod;
		bonusDirty=TRUE;
		if ((score[currentPlayer]+bonusMod)/10000 > score[currentPlayer]/10000)
		{	lives[currentPlayer]++; livesDirty[currentPlayer]=TRUE;	}
		score[currentPlayer]+=bonusMod;
		scoreDirty[currentPlayer]=TRUE;
		if (bonus[currentPlayer]<0)
		{
			bonus[currentPlayer]=0;
			stateVar=1;
			soDoneLevel->stop();
		}
	}
	else if (stateVar==1)
	{
		Sprite::cls();
		stateVar=2;
	}
	else if (stateVar>=2)
	{
		stateVar++;
		if (stateVar==35)
		{
			setLevel(currentPlayer,level[currentPlayer]+1);
			Sprite::cls();
			buildLevelSprite(currentPlayer);
		}
		if (stateVar==60)
		{
			stateVar=0;
			state=PLAYING;
		}
	}
	break;

case KEYREDEFINITION:
	if (stateVar>4) stateVar++;
	if (stateVar==50)
	{
		if (options.persistentKeys) writeKeys();
		Sprite::cls();
		state=DRAWINGMENU;
		stateVar=0;
	}
	break;
}
return TRUE;
}

BOOL Game::draw()
{

/*	static int speedCount;
	if (gameSpeed==2)
		if (speedCount++%3==0) return FALSE;*/
static int c; // temp loop counter
static int e; // temp elavator pos
static int blackCount; // temp timer for clearing menu dust
static int a; // temp animation store
static int x,y; // temp loop counters

switch(state)
{
case DONELEVEL:
	if (stateVar>0 && stateVar<35) break;
case PLAYING:
case PLAYERDEATH:

DrawLevel:

	RECT blackPatch;
	blackPatch.left=0; blackPatch.right=200*magnify;
	blackPatch.top=121*magnify; blackPatch.bottom=162*magnify;

	if (fullScreen)
	{
		if (blackCount++%21==0)
		{
			spFrontend->draw(0,0,blackPatch);
			spFrontend->draw(200,0,blackPatch);
			spFrontend->draw(400,0,blackPatch);
			spFrontend->draw(600,0,blackPatch);
		}
	}

	for(c=0; c<noOfPlayers; c++)
	{
		if (scoreDirty[c])
		{
			drawNumber(1, 1, 56+c*68-2, 0, (c==currentPlayer)?FALSE:TRUE);
			drawNumber(score[c], 6, 56+c*68, 0, (c==currentPlayer)?FALSE:TRUE);
			scoreDirty[c]=FALSE;
		}
		if (livesDirty[c])
		{
			RECT rect;
			rect.left=76*magnify; rect.right=82*magnify;
			rect.top=18*magnify; rect.bottom=20*magnify; // size of token
			RECT token;
			token.left=48*magnify; token.right=54*magnify;
			token.top=19*magnify; token.bottom=21*magnify;
			for(int l=0;l<8;l++)
				spNumbers->draw(spFrontend,(54+c*68+l*8)*magnify, 10*magnify,rect);
			for(l=0; l<((lives[c]>8)?8:lives[c]); l++)
				spNumbers->draw(spFrontend,(54+c*68+l*8)*magnify, 10*magnify, token);
			livesDirty[c]=FALSE;
		}
	}
	if (playerDirty)
	{
		drawNumber(currentPlayer+1, 1, 54, 16, FALSE);
		playerDirty=FALSE;
	}
	if (levelDirty)
	{
		drawNumber(level[currentPlayer]+1, 2, 128, 16, FALSE);
		levelDirty=FALSE;
	}
	if (bonusDirty)
	{
		drawNumber(bonus[currentPlayer], 4, 204, 16, FALSE);
		bonusDirty=FALSE;
	}
	if (timeDirty)
	{
		drawNumber(time, 3, 290, 16, FALSE);
		timeDirty=FALSE;
	}

	RECT blank;
	blank.left=0; blank.right=16*magnify;
	blank.top=121*magnify; blank.bottom=129*magnify;

	if (runner.y<40)
	{
		spFrontend->draw((runner.x-4)*2*magnify+xmod,25*magnify+ymod,blank);
		spFrontend->draw((runner.x+4)*2*magnify+xmod,25*magnify+ymod,blank);
	}

	if (chuckie.y<40)
	{
		blank.right=24*magnify;
		if (chuckie.dir==1)
		 // left
			spFrontend->draw((chuckie.x-6)*2*magnify+xmod,25*magnify+ymod,blank);
		else
			// right
			spFrontend->draw((chuckie.x-2)*2*magnify+xmod,25*magnify+ymod,blank);
	}

	spLevel->draw(xmod, 32*magnify+ymod);
	RECT levelTop;
	levelTop.left=0; levelTop.right=320*magnify;
	levelTop.top=0; levelTop.bottom=32*magnify;
	spFrontend->draw(xmod, ymod, levelTop);
	RECT cage;
	cage.left=0; cage.right=48*magnify; cage.top=18*magnify; cage.bottom=66*magnify;
	spNumbers->draw(xmod,28*magnify+ymod,cage);

	RECT beak;
	beak.left=0; beak.right=8*magnify;
	beak.top=121*magnify; beak.bottom=129*magnify;

	if (chuckie.caged) spFrontend->draw(34*magnify+xmod,44*magnify+ymod,beak);

	RECT elevator;

	if (map[cp].elevator!=0)
	{
		blank.left=0; blank.right=32*magnify; blank.top=121*magnify; blank.bottom=129*magnify;
		spFrontend->draw((map[cp].elevator*16-16)*magnify+xmod, 248*magnify+ymod, blank);
		blank.left=0; blank.right=32*magnify; blank.top=121*magnify; blank.bottom=129*magnify;
		spFrontend->draw((map[cp].elevator*16-16)*magnify+xmod, 26*magnify+ymod, blank);
		elevator.left=56*magnify; elevator.right = 76*magnify; elevator.top=18*magnify; elevator.bottom=22*magnify;
		spNumbers->draw((map[cp].elevator*16-10)*magnify+xmod, (elevatorPos+26)*magnify+ymod, elevator);
		e=elevatorPos-80;
		if (e<0) e+=226;
		elevator.left=56*magnify; elevator.right = 76*magnify; elevator.top=18*magnify; elevator.bottom=22*magnify;
		spNumbers->draw((map[cp].elevator*16-10)*magnify+xmod, (e+26)*magnify+ymod, elevator);
	}

	if (state!=STARTINGGAME)
	{
		RECT rect;
		rect.left=0; rect.right=16*magnify;
		int animMult=16*magnify;
		if (runner.anim==0 || runner.anim==2) a=0;
		if (runner.anim==1) a=1;
		if (runner.anim==3) a=2;
		if (runner.squirm!=-1) a=upPressed?2:0;
		if ((!runner.onLadder && !runner.jumpFromLadder) || runner.squirm!=-1)
		{
			if (runner.squirm==1 || (runner.squirm==-1 && runner.dir==1))
				rect.top=0;//spRunnerR[a]->draw(runner.x*2*magnify+xmod,(runner.y+24)*magnify+ymod);
			else
				rect.top=48*magnify;//spRunnerL[a]->draw(runner.x*2*magnify+xmod,(runner.y+24)*magnify+ymod);
		}
		else
		{
			rect.top=96*magnify; animMult=18*magnify;
//			spRunnerV[a]->draw(runner.x*2*magnify+xmod,(runner.y+24)*magnify+ymod);
		}
		rect.top+=a*animMult;
		rect.bottom=rect.top+animMult;
		int runnerX=runner.x;
		if (runnerX<0) runnerX=0;
		if (runnerX>152) runnerX=152;
		spRunner->draw(runnerX*2*magnify+xmod,(runner.y+24)*magnify+ymod,rect);
	}

	RECT swanRect;

	for(c=0; c<8; c++)
		if (map[cp].swans[c].x!=-1)
		{
			swanRect.left = 0; swanRect.right= 28*magnify;
			int swanX=map[cp].swans[c].x*2*magnify+xmod;
			int swanY=(map[cp].swans[c].y+20)*magnify+ymod;

			switch(map[cp].swans[c].dir)
			{
			case 0:
			case 2:
				swanRect.top=40*magnify;
				if (map[cp].swans[c].y%8<4) {swanRect.top+=100*magnify;swanX-=14*magnify;}
				break;

			case 1:
				if (map[cp].swans[c].pecking==1 || map[cp].swans[c].pecking==3)
					swanRect.top=60*magnify;

				else if (map[cp].swans[c].pecking==2)
					swanRect.top=80*magnify;

				else
					swanRect.top=((map[cp].swans[c].x%8)==0?0:20)*magnify;

				break;

			case 3:
				swanX-=14*magnify;
				if (map[cp].swans[c].pecking==1 || map[cp].swans[c].pecking==3)
					swanRect.top=160*magnify;

				else if (map[cp].swans[c].pecking==2)
					swanRect.top=180*magnify;

				else
					swanRect.top=((map[cp].swans[c].x%8)==0?100:120)*magnify;

				break;
			}
			swanRect.bottom=swanRect.top+20*magnify;
			spSwans->draw(swanX,swanY,swanRect);
		}

	RECT chuckieRect;
	chuckieRect.left=0; chuckieRect.right=32*magnify;
	chuckieRect.top=(((chuckie.dir!=0)?48:0)+((chuckie.anim<4)?24:0))*magnify;
	chuckieRect.bottom=chuckieRect.top+24*magnify;
	spChuckie->draw(chuckie.x*2*magnify+xmod, (chuckie.y+24)*magnify+ymod,chuckieRect);

	break;

case MENU:
	if ((stateVar>=500 && stateVar<750) || stateVar>=1050) goto DrawLevel;
	if (stateVar>=750 && stateVar<1050)
	{
		RECT logo;
		logo.left=0; logo.right=320*magnify; logo.top=32*magnify; logo.bottom=62*magnify;
		spFrontend->draw(xmod,ymod,logo);
		drawText("KEYS",130,55,YELLOW);
		drawText("Up ..",50,80,CYAN);
		drawKey(moveUpKey,144,80,CYAN);
		drawText("Down ..",18,100,CYAN);
		drawKey(moveDownKey,144,100,CYAN);
		drawText("Left ..",18,120,CYAN);
		drawKey(moveLeftKey,144,120,CYAN);
		drawText("Right ..",2,140,CYAN);
		drawKey(moveRightKey,144,140,CYAN);
		drawText("Jump ..",18,160,CYAN);
		drawKey(jumpKey,144,160,CYAN);
		drawText("Hold .. 'P'",18,185,MAGENTA);
		drawText("Abort .. Escape",2,205,MAGENTA);
		drawText("Press",34,230,YELLOW);
		drawText("S",130,230,MAGENTA);
		drawText("to start",162,230,YELLOW);
		drawText("K",34,243,MAGENTA);
		drawText("to change keys",66,243,YELLOW);
		break;
	}

case DRAWINGMENU:
	if (stateVar==1 || stateVar>20)
	{
		RECT logo;
		logo.left=0; logo.right=320*magnify; logo.top=32*magnify; logo.bottom=62*magnify;
		spFrontend->draw(xmod,8*magnify+ymod,logo);
	}
	else if (stateVar==0)
	{
		RECT rect;
		rect.left=0; rect.right=320*magnify;
		rect.top=32*magnify; rect.bottom=47*magnify;
		spFrontend->draw(xmod,8*magnify+ymod,rect);
	}
	if (stateVar==3 || stateVar>20) drawText("HIGH SCORES",74,54,YELLOW);
	static int cc;
	static char hn[10];
	for(cc=0;cc<10;cc++)
	{
		if (stateVar==4+cc || stateVar>20)
		{
			drawText(itoa(cc+1,hn,10),cc==9?10:26,78+cc*12,GREEN);
			drawText(itoa(highScores[cc],hn,10),90,78+cc*12,GREEN);
			drawText(highNames[cc],190,78+cc*12,GREEN);
		}
	}
	if (stateVar==14 || stateVar>20) drawText("Press",34,230,YELLOW);
	if (stateVar==14 || stateVar>20) drawText("S",130,230,MAGENTA);
	if (stateVar==14 || stateVar>20) drawText("to start",162,230,YELLOW);
	if (stateVar==15 || stateVar>20) drawText("K",34,243,MAGENTA);
	if (stateVar==15 || stateVar>20) drawText("to change keys",66,243,YELLOW);
	break;

case STARTINGGAME:
	if (stateVar<4)
	{
		drawText("Get ready",90,130,YELLOW);
		static char sg[]="Player  "; sg[7]=currentPlayer+49;
		drawText(sg,98,155,CYAN);
	}
	if (stateVar>90) goto DrawLevel;
	break;

case GAMEOVER:
//  	if (stateVar==0)
 //	{
		static Sprite* b; b = new Sprite;
		b->setSize(192*magnify,50*magnify);
		b->draw(64*magnify+xmod,121*magnify+ymod);
		delete b;
  //	}
	drawText("GAME OVER",90,130,CYAN);
	drawText("Player ",98,155,CYAN);
	static char jj[2]; jj[1]='\0';
	jj[0]=currentPlayer+49;
	drawText(jj,210,155,CYAN);
	break;

case KEYREDEFINITION:
	drawText("K E Y",121,18,YELLOW);
	drawText("S E L E C T I O N",26,43,YELLOW);
	drawText("Up ..",50,80,MAGENTA);
	if (stateVar>0)
	{
		drawKey(moveUpKey,144,80,MAGENTA);
		drawText("Down ..",18,100,MAGENTA);
	}
	if (stateVar>1)
	{
		drawKey(moveDownKey,144,100,MAGENTA);
		drawText("Left ..",18,120,MAGENTA);
	}
	if (stateVar>2)
	{
		drawKey(moveLeftKey,144,120,MAGENTA);
		drawText("Right ..",2,140,MAGENTA);
	}
	if (stateVar>3)
	{
		drawKey(moveRightKey,144,140,MAGENTA);
		drawText("Jump ..",18,160,MAGENTA);
	}
	if (stateVar>4)
		drawKey(jumpKey,144,160,MAGENTA);
	break;

case PLAYERPROMPT:
	static char p[2];p[0]=noOfPlayers+48;p[1]='\0';
	if (stateVar<4) drawText("How many players?",10,130,YELLOW);
	if (stateVar>0) drawText(p,298,130,YELLOW);
	break;

case HIGHSCORE:
	RECT top;
	top.left=0; top.right=320*magnify;
	top.top=0; top.bottom=32*magnify;
	spFrontend->draw(xmod,ymod,top);
	drawText("HIGH SCORES",74,54,YELLOW);
	int hscc;
	char hshn[10];
	for(hscc=0;hscc<10;hscc++)
	{
		drawText(itoa(hscc+1,hshn,10),hscc==9?10:26,78+hscc*12,GREEN);
		drawText(itoa(highScores[hscc],hshn,10),90,78+hscc*12,GREEN);
		if (hscc!=ranking)
			drawText(highNames[hscc],190,78+hscc*12,GREEN);
		else
		{
			drawText(">",174,78+hscc*12,YELLOW);
			drawText(highNames[hscc],190,78+hscc*12,YELLOW);
		}
	}
	drawText("ENTER YOUR NAME",42,215,YELLOW);
	static char hs[]="Player x";
	hs[7]=currentPlayer+49;
	drawText(hs,98,230,MAGENTA);
	break;
}

	if (showFps>0) showFps--;
	if (showFps>0)
	{
		static Sprite* fps;
		fps=new Sprite;
		fps->setSize(164*magnify,30*magnify);
		if (showFps>2)
		{
			fps->lock();
			fps->drawRect(0,0,164*magnify-1,30*magnify-1,255);
			fps->unlock();
			fps->draw(100*magnify+xmod,100*magnify+ymod);
			drawText("fps = ",105,110,YELLOW);
			char cfps[5];
			itoa(options.framesPerSecond,cfps,10);
			drawText(cfps,192,110,YELLOW);
		}
		else fps->draw(100*magnify+xmod,100*magnify+ymod); // clear rectangle after timeout
		delete fps;
	}

	return TRUE;
}

BOOL Game::restore()
{
	for(int c=0; c<4; c++)
	{
		scoreDirty[c]=TRUE;
		livesDirty[c]=TRUE;
	}
	playerDirty=TRUE; levelDirty=TRUE; bonusDirty=TRUE; timeDirty=TRUE;
//	if (fullScreen)
	{
		Sprite* b=new Sprite;
		b->setSize(800,600);
		b->draw(0,0);
		Sprite::flipScreens();
		b->draw(0,0);
		delete b;
	}
	spLevel->setSize(320*magnify,224*magnify);
	buildLevelSprite(currentPlayer);
	return TRUE;
}

void Game::closeDown()
{
}

void Game::mouseMove(int,int)
{
}

void Game::mouseLeftDown(int,int)
{
}

void Game::mouseLeftUp(int,int)
{
}

void Game::mouseRightDown(int,int)
{
}

void Game::mouseRightUp(int,int)
{
}

void Game::keyDown(unsigned char k)
{

#ifdef easyCheats
if (k==VK_F2)
	if (invulnerable)
		{invulnerable=FALSE;showMessage("Invincible mode off");}
	else
		{
			invulnerable=TRUE;
			showMessage("Invincible mode on");
			options.cheating=TRUE;
			options.reasons.invulnerable=TRUE;
		}
#endif

static unsigned int invulCount=0;
static char invulChar[]="NOO";
if (k!=invulChar[invulCount])
	invulCount=0;
else
{
	invulCount++;
	if (invulCount==strlen(invulChar))
	{
		if (invulnerable)
			{invulnerable=FALSE;showMessage("Invincible mode off");}
		else
			{
				invulnerable=TRUE;
				showMessage("Invincible mode on");
				options.cheating=TRUE;
				options.reasons.invulnerable=TRUE;
			}
	}
}

static unsigned int levJumpCount=0;
static char levJumpChar[]="SMUDGE";
if (k!=levJumpChar[levJumpCount])
	levJumpCount=0;
else
{
	levJumpCount++;
	if (levJumpCount==strlen(levJumpChar))
	{
		if (levJump)
			{levJump=FALSE;showMessage("Level jump off");}
		else
			{
				levJump=TRUE;
				showMessage("Level jump on");
				options.cheating=TRUE;
				options.reasons.levelJump=TRUE;
			}
	}
}

if (k==pauseKey)
{
	paused=!paused;
	soWalk->stop(); soLadder->stop();
	soDoneLevel->stop(); soJump->stop();
	soFall->stop(); soDeath->stop();
}
if (k==16) shiftPressed=TRUE;
if (k==(char)1)
{
	Sprite::cls();
	state=DRAWINGMENU;
	stateVar=0;
	soWalk->stop();
	soJump->stop();
	soFall->stop();
	soDeath->stop();
	soDoneLevel->stop();
	soLadder->stop();
}

switch(state)
{
case PLAYING:
#ifdef easyCheats
	if (k==VK_F1)
	{
		setLevel(currentPlayer,level[currentPlayer]+1);
		options.cheating=TRUE;
		options.reasons.levelJump=TRUE;
	}
	if (k==VK_F3)
	{
		chuckie.caged=TRUE;
		chuckie.dir=0;
		chuckie.x=4;
		chuckie.y=20;
		options.cheating=TRUE;
		options.reasons.cageChuckie=TRUE;
	}
#else
	if (k==VK_F1 && levJump)
	{
		setLevel(currentPlayer,level[currentPlayer]+1);
		options.cheating=TRUE;
		options.reasons.levelJump=TRUE;
	}
	if (k==VK_F3 && (levJump||invulnerable))
	{
		chuckie.caged=TRUE;
		chuckie.dir=0;
		chuckie.x=4;
		chuckie.y=20;
		options.cheating=TRUE;
		options.reasons.cageChuckie=TRUE;
	}
#endif
	break;

case MENU:
case DRAWINGMENU:
	if (k=='k' || k=='K')
	{
		Sprite::cls();
		state=KEYREDEFINITION;
		stateVar=0;
	}
	if (k=='S' || k=='s')
	{
		Sprite::cls();
		state=PLAYERPROMPT;
		stateVar=0;
	}
	break;

case PLAYERPROMPT:
	if (stateVar==0 && k>='1' && k<='4')
	{
		static int c;
		noOfPlayers=k-48;
		stateVar=1;
		for(c=0;c<noOfPlayers;c++)
			level[c]=options.levelJump;
	}
	break;

case KEYREDEFINITION:
	if (stateVar==4 && k!=moveUpKey && k!=moveDownKey && k!=moveLeftKey && k!=moveRightKey) {jumpKey=k;stateVar++;}
	if (stateVar==3 && k!=moveUpKey && k!=moveDownKey && k!=moveLeftKey) {moveRightKey=k;stateVar++;}
	if (stateVar==2 && k!=moveUpKey && k!=moveDownKey) {moveLeftKey=k;stateVar++;}
	if (stateVar==1 && k!=moveUpKey) {moveDownKey=k;stateVar++;}
	if (stateVar==0) {moveUpKey=k;stateVar++;}
	break;

case HIGHSCORE:
	if (stateVar>0)
	{
		if (k==VK_RETURN) {stateVar=149;state=GAMEOVER;Sprite::cls();if (options.persistentScores) writeScores();}
		if (k>=32 && k<=128 && stateVar<9) {highNames[ranking][stateVar-1]=shiftPressed?k:k+32;stateVar++;}
		if (k==8 && stateVar>1) {highNames[ranking][stateVar-2]=' ';stateVar--;}
	}
	break;
}
}

void Game::keyUp(unsigned char k)
{
	if (k==16) shiftPressed=FALSE;
}

void Game::drawNumber(int num, int length, int x, int y, BOOL inverted)
{
	RECT rect;
	for(int c=0; c<length; c++)
	{
		rect.top=inverted?magnify*9:0; rect.bottom=rect.top+9*magnify;
		rect.left=(num%10)*10*magnify; rect.right=rect.left+10*magnify;
		spNumbers->draw(spFrontend, (x+(length-c-1)*10)*magnify, y*magnify, rect);
		num/=10;
	}
}

BOOL Game::setSize(BOOL fs, int width, int height)
{
	if (fs)
	{
		if (width!=800 || height!=600) return FALSE;
		Sprite* b=new Sprite;
		b->setSize(800,600);
		b->draw(0,0);
		Sprite::flipScreens();
		b->draw(0,0);
		delete b;
		magnify=2; loadSprites(magnify); restore();
		xmod=30; ymod=20;
		fullScreen=fs;
		return TRUE;
	}

	if (width%320!=0 || height%256!=0 || width/320!=height/256) return FALSE;
	magnify = width/320;
	Sprite* b=new Sprite;
	b->setSize(width,height);
	b->draw(0,0);
	xmod=0; ymod=0;
	delete b;
	loadSprites(magnify);
	restore();
	fullScreen=fs;
	return TRUE;
}

// SEED_VALUE used only on level 8 repeats after level 24
#define SEED_VALUE 1956

void Game::setLevel(int pl, int lev)
{
/*	if (map[cp].elevator!=0)
	{
		spBlank->draw((map[cp].elevator*16-16)*magnify+xmod, 248*magnify+ymod);
		spBlank->draw((map[cp].elevator*16)*magnify+xmod, 248*magnify+ymod);
		spBlank->draw((map[cp].elevator*16-16)*magnify+xmod, 26*magnify+ymod);
		spBlank->draw((map[cp].elevator*16)*magnify+xmod, 26*magnify+ymod);
	}*/

	RECT blank;
	blank.left=0; blank.right=32*magnify;
	blank.top=121*magnify; blank.bottom=153*magnify;

	if (maps[level[pl]%noOfMaps].elevator>0)
	{
		spFrontend->draw((map[pl].elevator*16-16)*magnify+xmod, 248*magnify+ymod, blank);
		Sprite::flipScreens();
		spFrontend->draw((map[pl].elevator*16-16)*magnify+xmod, 248*magnify+ymod, blank);
	}

	memcpy(&(map[pl]), &(maps[lev%noOfMaps]), sizeof(Map));
	level[pl]=lev;
	runner.x=map[pl].runner.x;
	runner.y=map[pl].runner.y;
	runner.dir=1;
	runner.anim=0;
	runner.onLadder=FALSE;
	runner.jumping=FALSE;
	runner.yvel=0;
	runner.xvel=0;
	runner.jumpFromLadder=FALSE;
	runner.minijump=FALSE;
	runner.onElevator=FALSE;
	runner.fallNudge=0;
	runner.squirm=-1;
	runner.byebye=0;

	chuckie.caged=(lev>=noOfMaps)?FALSE:TRUE;
	chuckie.x=4;
	chuckie.y=20;
	chuckie.dir=0;
	chuckie.xvel=0;
	chuckie.yvel=0;

	bonus[pl]=(lev<9)?(lev+1)*1000:9000;
	time=900-(level[pl]/16)*100; if (time<100) time=100;

	haltBonus=0;
	playerDirty=TRUE; timeDirty=TRUE; levelDirty=TRUE; bonusDirty=TRUE;
	livesDirty[pl]=TRUE; scoreDirty[pl]=TRUE;

	elevatorPos=223;

	for(int c=0; c<8; c++)
		if ((lev<noOfMaps || lev>noOfMaps*2-1) &&
			map[pl].swans[c].x!=-1 && (c<4 || lev>=noOfMaps*3) )
		{
			map[pl].swans[c].x = (map[pl].swans[c].x-1)*8;
			map[pl].swans[c].y = (27-map[pl].swans[c].y)*8;
		}
		else
			map[pl].swans[c].x=-1;

	buildLevelSprite(pl);
	if (lev>24) srand(9573); else srand(12753);
	if (level[pl]>23 && level[pl]%8==7) srand(SEED_VALUE);
}

void Game::die()
{
	soWalk->stop();
	soLadder->stop();
	soJump->stop();
	soFall->stop();
	if (soDeath!=NULL) soDeath->play();
	state=PLAYERDEATH;
	stateVar=0;
}

void Game::drawText(char* b, int x, int y, MYCOL col)
{
	RECT src;
	for(unsigned int cc=0; cc<strlen(b); cc++)
	{
		src.left=(((int)(b[cc])-32)%20)*16*magnify;
		src.top=(((int)(b[cc])-32)/20)*8*magnify;
		src.right=src.left+16*magnify;
		src.bottom=src.top+8*magnify;
		spFont[col]->draw((x+cc*16)*magnify+xmod,y*magnify+ymod,src);
	}
}

void Game::sendMessage(GAMEMESSAGE gm, int val)
{
	switch(gm)
	{
	case USEJOYSTICK:
		options.useJoystick=(BOOL)val;
		break;
	case SOUNDMODE:
		options.soundMode=val;
		Sound::setSoundMode(val);
		break;
	case PERSISTENTSCORES:
		options.persistentScores=val==1?TRUE:FALSE;
		if (val==1) writeScores();
		break;
	case PERSISTENTKEYS:
		options.persistentScores=val==1?TRUE:FALSE;
		if (val==1) writeKeys();
		break;
	case SETFPS:
		if (val>1000) val=1000; if (val<1) val=1;
		options.framesPerSecond=val;
		char fps[100];
		WritePrivateProfileString("Chuckie","FramesPerSecond",itoa(val,fps,10),iniFile);
		timePerFrame = 1000/val;
		showFps=val;
		if (val<40) {options.cheating=TRUE;options.reasons.tooSlow=TRUE;}
		break;
	case EXTENDEDJUMPS:
		options.extendedJumps=val==1?TRUE:FALSE;
		if (options.extendedJumps==TRUE)
		{
			options.cheating=TRUE;
			options.reasons.extJumps=TRUE;
		}
		WritePrivateProfileString("Chuckie","ExtendedJumps",val==1?"1":"0",iniFile);
		break;
	}
}

int Game::matrix(int x, int y)
{
	if (x>=0 && x<20 && y>=0 && y<27)
	{
		int res=map[currentPlayer].matrix[x][y];
		if (res!=INV_PLATFORM)
			return res;
		else
			return PLATFORM;
	}
	return BLANK;
}

BOOL Game::isPlatform(int x, int y)
{
	if (x<0 || x>=20 || y<0 || y>=27) return FALSE;
	int res=map[currentPlayer].matrix[x][y];
	if (res==PLATFORM || res==SOLID_LADDER || res==INV_PLATFORM
		|| (res==LADDER && ( matrix(x-1,y)==PLATFORM || matrix(x+1,y)==PLATFORM)))
		return TRUE;
	return FALSE;
}

BOOL Game::isLadder(int x, int y)
{
	if (x<0 || x>=20 || y<0 || y>=27) return FALSE;
	int res=map[currentPlayer].matrix[x][y];
	if (res==LADDER || res==SOLID_LADDER)
		return TRUE;
	return FALSE;
}

void Game::setMatrix(int x, int y, int sq)
{
	if (x>=0 && x<20 && y>=0 && y<27) map[currentPlayer].matrix[x][y]=sq;
}

void Game::resetLevel(int pl)
{
	runner.x=map[pl].runner.x;
	runner.y=map[pl].runner.y;
	runner.dir=1;
	runner.anim=0;
	runner.onLadder=FALSE;
	runner.jumping=FALSE;
	runner.yvel=0;
	runner.xvel=0;
	runner.jumpFromLadder=FALSE;
	runner.minijump=FALSE;
	runner.onElevator=FALSE;
	runner.fallNudge=0;
	runner.squirm=-1;
	runner.byebye=0;
	time=900-(level[pl]/16)*100; if (time<100) time=100;
	haltBonus=0;
	playerDirty=TRUE; timeDirty=TRUE; levelDirty=TRUE; bonusDirty=TRUE;
	livesDirty[pl]=TRUE; scoreDirty[pl]=TRUE;

	chuckie.caged=(level[pl]>=noOfMaps)?FALSE:TRUE;
	chuckie.x=4;
	chuckie.y=20;
	chuckie.dir=0;
	chuckie.xvel=0;
	chuckie.yvel=0;

	elevatorPos=223;

	for(int c=0; c<8; c++)
		if ((level[pl]<noOfMaps || level[pl]>noOfMaps*2-1) &&
			(maps[level[pl]%noOfMaps].swans[c].x!=-1) && (c<4 || level[pl]>=(noOfMaps*3)))
		{
			map[pl].swans[c].x = (maps[level[pl]%noOfMaps].swans[c].x-1)*8;
			map[pl].swans[c].y = (27-maps[level[pl]%noOfMaps].swans[c].y)*8;
			map[pl].swans[c].dir=1;
			map[pl].swans[c].pecking=0;
		}
		else
			map[pl].swans[c].x=-1;
	if (level[pl]>24) srand(12845); else srand(18903);
	if (level[pl]>23 && level[pl]%8==7) srand(SEED_VALUE);
}

void Game::pickUp(int x, int y)
{
	RECT blank;
	blank.left=0; blank.right=16*magnify;
	blank.top=121*magnify; blank.bottom=129*magnify;

	if (matrix(x,y)==EGG)
	{
		setMatrix(x,y,BLANK);
		soEgg->play(runner.x*20000/156-10000);
		if ((score[currentPlayer]+100)/10000 > score[currentPlayer]/10000)
		{	lives[currentPlayer]++; livesDirty[currentPlayer]=TRUE;	}
		score[currentPlayer]+=100;
		scoreDirty[currentPlayer]=TRUE;
		int eggCount=0;
		for(int xx=0; xx<20; xx++)
			for(int yy=0; yy<27; yy++)
				if (matrix(xx,yy)==EGG) eggCount++;
		if (eggCount==0)
		{
			soWalk->stop();
			soLadder->stop();
			soJump->stop();
			soFall->stop();
			soDoneLevel->playLooped();
			state=DONELEVEL;stateVar=0;
		}
		spFrontend->draw(spLevel,x*16*magnify,y*8*magnify, blank);
	}
	if (matrix(x,y)==GRAIN)
	{
		setMatrix(x,y,BLANK);
		soGrain->play(runner.x*20000/156-10000);
		if ((score[currentPlayer]+50)/10000 > score[currentPlayer]/10000)
		{	lives[currentPlayer]++; livesDirty[currentPlayer]=TRUE;	}
		score[currentPlayer]+=50;
		scoreDirty[currentPlayer]=TRUE;
		haltBonus=15;
		spFrontend->draw(spLevel,x*16*magnify,y*8*magnify, blank);
	}
}

void Game::drawKey(unsigned char k, int x, int y, MYCOL col )
{
	char jj[2]=".";
	switch(k)
	{
	case VK_TAB:
		drawText("Tab",x,y,col);
		break;
	case VK_SPACE:
		drawText("Space",x,y,col);
		break;
	case VK_RETURN:
		drawText("Return",x,y,col);
		break;
	case VK_BACK:
		drawText("Backspace",x,y,col);
		break;
	case VK_INSERT:
		drawText("Insert",x,y,col);
		break;
	case VK_DELETE:
		drawText("Delete",x,y,col);
		break;
	case VK_END:
		drawText("End",x,y,col);
		break;
	case VK_PRIOR:
		drawText("Page up",x,y,col);
		break;
	case VK_NEXT:
		drawText("Page down",x,y,col);
		break;
	case VK_HOME:
		drawText("Home",x,y,col);
		break;
	case VK_CAPITAL:
		drawText("Caps lock",x,y,col);
		break;
	case VK_LEFT:
		drawText("Left arrow",x,y,col);
		break;
	case VK_RIGHT:
		drawText("Right arrow",x,y,col);
		break;
	case VK_UP:
		drawText("Up arrow",x,y,col);
		break;
	case VK_DOWN:
		drawText("Down arrow",x,y,col);
		break;
	case VK_SHIFT:
		drawText("Shift",x,y,col);
		break;
	case VK_CONTROL:
		drawText("Control",x,y,col);
		break;
	default:
		if (k==186) k=59; // ';'
		if (k==187) k=61; // '='
		if (k==188) k=44; // ','
		if (k==189) k=45; // '-'
		if (k==190) k=46; // '.'
		if (k==191) k=47; // '/'
		if (k==192) k=39; // '''
		if (k==219) k=91; // '['
		if (k==220) k=92; // '\'
		if (k==221) k=93; // ']'
		if (k==222) k=35; // '#'
		if (k==223) k=96; // '`'
		drawText("' '",x,y,col);
		if (k>=65 && k<=90) k+=32;
		jj[0]=k;
		drawText(jj,x+16,y,col);
		break;
	}
}

void Game::writeScores()
{
	for(int cc=0; cc<10; cc++)
	{
		char n[50],s[20];
		strcpy(n,"Highname ");n[8]=(char)cc+48;
		s[0]='"';strcpy(s+1,highNames[cc]);s[9]='"';s[10]='\0';
		WritePrivateProfileString("Chuckie",n,s,iniFile);
		strcpy(n,"Highscore ");n[9]=(char)cc+48;
		WritePrivateProfileString("Chuckie",n,itoa(highScores[cc],s,10),iniFile);
	}
}

void Game::writeKeys()
{
	char k[20];
	WritePrivateProfileString("Chuckie","LeftKey",itoa((int)moveLeftKey,k,10),iniFile);
	WritePrivateProfileString("Chuckie","RightKey",itoa((int)moveRightKey,k,10),iniFile);
	WritePrivateProfileString("Chuckie","UpKey",itoa((int)moveUpKey,k,10),iniFile);
	WritePrivateProfileString("Chuckie","DownKey",itoa((int)moveDownKey,k,10),iniFile);
	WritePrivateProfileString("Chuckie","JumpKey",itoa((int)jumpKey,k,10),iniFile);
}

void Game::buildLevelSprite(int pl)
{

	RECT platform, ladder, egg, grain, blank;
	platform.left=0; platform.right=16*magnify; platform.top=0; platform.bottom=8*magnify;
	ladder.left=0; ladder.right=16*magnify; ladder.top=8*magnify; ladder.bottom=16*magnify;
	egg.left=0; egg.right=16*magnify; egg.top=16*magnify; egg.bottom=24*magnify;
	grain.left=0; grain.right=16*magnify; grain.top=24*magnify; grain.bottom=32*magnify;
	blank.left=0; blank.right=16*magnify; blank.top=121*magnify; blank.bottom=129*magnify;

	for(int y=0; y<27; y++)
		for(int x=0; x<20; x++)
			switch(map[pl].matrix[x][y])
			{
case INV_PLATFORM :
case BLANK : spFrontend->draw(spLevel,x*16*magnify,y*8*magnify, blank); break;
case PLATFORM : spBlocks->draw(spLevel,x*16*magnify,y*8*magnify,platform); break;
case SOLID_LADDER :
case LADDER : spBlocks->draw(spLevel,x*16*magnify,y*8*magnify,ladder); break;
case EGG : spBlocks->draw(spLevel,x*16*magnify,y*8*magnify,egg); break;
case GRAIN : spBlocks->draw(spLevel,x*16*magnify,y*8*magnify,grain); break;
			}

	blank.right=320*magnify;

	for(int x=0; x<20; x++)
		spFrontend->draw(x*16*magnify,27*8*magnify, blank);
}

BOOL Game::save(char* fname)
{
	char* fullname;
	char* EXT = ".egg";
	if (strlen(fname)<strlen(EXT) ||
		strcmp(fname+strlen(fname)-strlen(EXT), EXT))
	{
		fullname = new char[strlen(fname)+strlen(EXT)+1];
		sprintf(fullname, "%s%s\0", fname, EXT);
	}
	else
	{
		fullname = new char[strlen(fname)+1];
		sprintf(fullname, "%s\0", fname);
	}

	if (state!=PLAYING && state!=DONELEVEL && state!=STARTINGGAME
		&& state!=GAMEOVER && state!=PLAYERPROMPT)
		return FALSE;

	char *buffer = new char[15000]; // max possible + safety zone
	memset(buffer,'\0',15000);

	sprintf(buffer,"CHUCK%d",CURRENT_VERSION);
	memcpy(buffer+8,&options.check,4); // we store level checksum to avoid problems
						// with different level files (refuse to load if different)
	memcpy(buffer+12,&noOfPlayers,4);
	memcpy(buffer+16,&currentPlayer,4);
	memcpy(buffer+20,&state,4);
	memcpy(buffer+24,&stateVar,4);
	memcpy(buffer+28,&elevatorPos,4);
	memcpy(buffer+32,&paused,4);

	memcpy(buffer+36,&haltBonus,4);
	memcpy(buffer+40,&time,4);
	memcpy(buffer+44,&runner,sizeof(runner)); assert(sizeof(runner)==56);
	memcpy(buffer+100,&chuckie,sizeof(chuckie)); assert(sizeof(chuckie)==28);
	memcpy(buffer+128,map,sizeof(Map)*noOfPlayers); assert(sizeof(Map)==2300);

	char* ptr=buffer+128+sizeof(Map)*noOfPlayers;
	memcpy(ptr,score,noOfPlayers*4); ptr+=noOfPlayers*4;
	memcpy(ptr,lives,noOfPlayers*4); ptr+=noOfPlayers*4;
	memcpy(ptr,level,noOfPlayers*4); ptr+=noOfPlayers*4;
	memcpy(ptr,bonus,noOfPlayers*4); ptr+=noOfPlayers*4;

	int l=strlen(readProfileStr("LevelFile"))+1;
	memcpy(ptr,&l,4); ptr+=4;
	strcpy((char*)ptr,readProfileStr("LevelFile")); ptr+=l;

	// add file extension if not already present
/*	if (strlen(fname)>4)
	{
		if (strcmp(fname+strlen(fname)-4,".egg"))
			strcat(fname,".egg");
	}
	else
		strcat(fname,".egg");*/

	int handle = open(fullname,O_WRONLY|O_CREAT|O_TRUNC|O_BINARY,S_IREAD|S_IWRITE);
	if (handle==-1) {showMessage("Unable to save file");return FALSE;}

	write(handle,buffer,ptr-buffer);
	close(handle);
	delete buffer;
	delete[] fullname;
	return TRUE;
}

BOOL Game::load(char* fname)
{
	// add file extension if not already present
	if (strlen(fname)>4)
	{
		if (strcmp(fname+strlen(fname)-4,".egg"))
			strcat(fname,".egg");
	}
	else
		strcat(fname,".egg");

	int handle;
	if ((handle=open(fname, O_RDONLY|O_BINARY))==-1)
	{
		showMessage("Unable to get file handle.");
		showMessage(fname);
		return FALSE;
	}

	char header[9]; read(handle, header, 8); header[8]='\0';
	char version[9]; sprintf(version,"CHUCK%d\n",CURRENT_VERSION); version[8]='\0';

	if (strcmp(header,version))
	{
		char mess[200]; sprintf(mess, "Does not appear to be a Chuckie v%d.%02d save file",
			CURRENT_VERSION/100,CURRENT_VERSION%100);
		showMessage(mess);
		close(handle);
		return FALSE;
	}

	int check; read(handle, &check, 4);

	read(handle, &noOfPlayers,4);
	read(handle, &currentPlayer, 4);
	read(handle, &state, 4);
	read(handle, &stateVar, 4);
	read(handle, &elevatorPos, 4);
	read(handle, &paused, 4);

	read(handle, &haltBonus, 4);
	read(handle, &time, 4);

	read(handle, &runner, sizeof(runner));
	read(handle, &chuckie, sizeof(chuckie));

	read(handle, map, sizeof(Map)*noOfPlayers);
	read(handle, score, noOfPlayers*4);
	read(handle, lives, noOfPlayers*4);
	read(handle, level, noOfPlayers*4);
	read(handle, bonus, noOfPlayers*4);

	int l;
	read(handle,&l,4);
	char lev[300];
	read(handle,lev,l);
	lev[l]='\0';

	if (!loadLevels(lev))
		showMessage("Cannot load levels for this file.");

	if (check!=options.check)
	{
		showMessage("This save game used a different version of level.txt - cannot load.");
		close(handle);
		return FALSE;
	}

	Sprite::cls();
	buildLevelSprite(currentPlayer);
	for(int c=0;c<4;c++)
	{
		scoreDirty[c]=TRUE;
		livesDirty[c]=TRUE;
	}
	playerDirty=TRUE; levelDirty=TRUE; bonusDirty=TRUE; timeDirty=TRUE;

	close(handle);
	options.cheating=TRUE;
	options.reasons.gameLoad=TRUE;
	return TRUE;
}

unsigned long Game::countBits(unsigned long num)
{
	unsigned long res=0;
	while(num>0)
	{
		if (num&1==1) res++;
		num=num>>1;
	}
	return res;
}
