#ifndef DoneGame

#include "sprite.h"
#include "sound.h"
#include "filerdr.h"

const unsigned int CURRENT_VERSION = 110;
#define VERSION_STRING 1.1

class Game
{
public:
	enum GAMESTATE { MENU , PLAYING , DRAWINGMENU, CLEARINGMENU, DONELEVEL ,
		STARTINGGAME, PLAYERDEATH, GAMEOVER, KEYREDEFINITION, HIGHSCOREENTRY,
		PLAYERPROMPT, HIGHSCORE};
	enum GAMEMESSAGE { USEJOYSTICK , SOUNDMODE ,PERSISTENTSCORES, PERSISTENTKEYS,
		SETFPS, EXTENDEDJUMPS};
	struct OPTIONS {
		int soundMode;
		int levelJump;
		int startLives;
		BOOL persistentKeys, persistentScores, jumpExt, useJoystick;
		BOOL fastRedraws, extendedJumps;
		long framesPerSecond;
		BOOL cheating;
        int check;
		unsigned long topScore, topLevel;
		struct {
			BOOL startLevel, startLives, levelsChanged, levelJump, cageChuckie,
				invulnerable, tooSlow, gameLoad, extJumps;
		} reasons; // justify not giving competition code
	};
private:
	enum MYCOL { YELLOW,CYAN,MAGENTA,GREEN,BLUE };

	Sprite *spLevel;

	Sprite *spFont[5];
	Sprite *spChuckie;

	Sprite *spRunner, *spBlocks, *spNumbers, *spSwans;
	Sprite *spFrontend; // level top, chuckie logo, font, numbers and colours

	Sound *soEgg, *soGrain, *soDeath, *soWalk, *soJump, *soDoneLevel, *soLadder;
	Sound *soFall;

	BOOL scoreDirty[4], playerDirty, levelDirty, bonusDirty, timeDirty, livesDirty[4];

	struct Map {
		int matrix[20][27];
		int elevator;
		struct {
			int x,y;
			int dir;
			int pecking;
		} swans[8];
		POINT runner;
	};
	Map *maps, map[4];
	int noOfMaps;
	GAMESTATE state;
	OPTIONS options;
	unsigned int highScores[10];
	char highNames[10][10];
	int ranking;
	BOOL shiftPressed;
	BOOL invulnerable,levJump;
	struct {
		int x,y;
		int xvel, yvel;
		int dir;
		int anim;
		BOOL onLadder;
		BOOL jumping;
		BOOL minijump;
		BOOL onElevator;
		int fallNudge;
		int squirm; // if jumping onto ladder and holding down jump, keep displaying l/r sprite
		BOOL jumpFromLadder; // used to retail ladder graphic when jumping stringht up from ladder
        int byebye; // if trying to jump two across, one up make sure player bounces
	} runner;
	BOOL fullScreen;
	unsigned char moveLeftKey, moveRightKey, moveUpKey, moveDownKey;
	unsigned char jumpKey, pauseKey, quitKey;
	BOOL leftPressed, rightPressed, upPressed, downPressed, jumpPressed;
	BOOL fire1Pressed,fire2Pressed,fire3Pressed;
	int elevatorPos;
	int magnify;    // global magnify for all graphics
	int xmod, ymod; // offset to begin graphics
	int gameSpeed; // global speed correction
	long timePerFrame; // in milliseconds
	long nextFrameTime; // windows tick count to begin next frame
	int showFps; // counter to display fps
	BOOL paused;
	unsigned int score[4];
	int level[4], lives[4], bonus[4];
	int time, haltBonus;
	int currentPlayer, noOfPlayers;
	char* iniFile;

	int stateVar;

	struct {
		BOOL caged;
		int dir;
		int x,y;
		int anim;
		int xvel, yvel;
	} chuckie;

	void drawNumber(int num,int length,int x,int y,BOOL inverted);
	void setLevel(int pl, int lev); // initialises player pl with level lev
	void die();
	int getymod(int vel);
	void drawText(char*,int x,int y,MYCOL col);
	int matrix(int,int);
	BOOL isPlatform(int,int);
	BOOL isLadder(int,int);
	BOOL isEgg(int,int);
    BOOL isGrain(int,int);
	void setMatrix(int,int,int);
	void resetLevel(int pl); // restores level state after death
	void pickUp(int x,int y); // test for egg/grain and collect
	void drawKey(unsigned char,int,int,MYCOL); // used in key redefinition
	void writeKeys();
	void writeScores();
	void buildLevelSprite(int);

	unsigned long countBits(unsigned long);

public:
	Game();
	~Game();

	BOOL init(BOOL,int,int);
	BOOL restore();
	void closeDown();
	BOOL iterate();
	BOOL draw();
	GAMESTATE getState() {return state;}
	void sendMessage(GAMEMESSAGE,int);
	OPTIONS getOptions() {return options;}

	BOOL setSize(BOOL,int,int); // inform Game of new window size. return TRUE if acceptable

	void loadSprites(int);
	void setIniFile(char* i) {iniFile=i;}

	BOOL save(char*);
	BOOL load(char*);
    BOOL loadLevels(char*);

	void mouseMove(int,int);       // handle mouse events
	void mouseLeftDown(int,int);   //
	void mouseLeftUp(int,int);     //
	void mouseRightDown(int,int);  //
	void mouseRightUp(int,int);    //
	void keyDown(unsigned char);      // handle key presses
	void keyUp(unsigned char);        //
};

#define DoneGame
#endif
