// main.cpp
// Contains basic Windows code to initialise a full screen window and the DirectX
// API, regularly call the iterate() and draw() members of the Simulation class,
// and pass on mouse clicks and keyboard strokes to the Simulation class.

// The code here is mostly taken from the program ddex4.cpp from the tutorial in
// the Microsoft DirectX SDK and used with the implicit permission of Microsoft.

#define NAME "Chuckie"
#define TITLE "Chuckie Egg"

#ifndef DoneWindows
//#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <windowsx.h>
#endif
#define DoneWindows

#ifndef DoneDDraw
#include <ddraw.h>
#endif
#define DoneDDraw

#include <stdio.h>
#include <string.h>
#include "resource.h"
#include "main.h"
#include "sprite.h"
#include "sound.h"
#include <errno.h>
#include <io.h>
#include <fcntl.h>
#include <sys\stat.h>
#include <stdlib.h>
#include "game.h"
#include <assert.h>
#include <sstream>

void setCursorPos(int x,int y) { SetCursorPos(x,y); }

BOOL bActive;

Game game;

void restoreAll( void );

HWND ghwnd;
HCURSOR hCursor;
HACCEL hAccel;
HINSTANCE ghinstance;
char appDir[256], iniFile[256], exeFile[256];
char commandLine[256];
Game::OPTIONS options;
int fps;
int noOfLevelFiles;
char* levelFiles[256];

BOOL useJoystick;
int soundMode;
BOOL fullScreen;
int scrnWidth, scrnHeight, windowWidth, windowHeight;
int maxMagnify;

void showError(int val)
{
	// converts a DirectX error code into a text message
	HRESULT ddrval=(HRESULT)val;
	switch( ddrval)
	{
	case DDERR_INVALIDPARAMS : oneMessage("invalid params");break;
	case DDERR_GENERIC : oneMessage("generic error");break;
	case DDERR_INVALIDRECT : oneMessage("invalid rect");break;
	case DDERR_WASSTILLDRAWING : oneMessage("still drawing");break;
	case DDERR_SURFACELOST : oneMessage("surface lost");break;
	case DDERR_INVALIDOBJECT : oneMessage("invalied object");break;
	case DDERR_OUTOFMEMORY : oneMessage("out of memory");break;
	case DDERR_SURFACEBUSY : oneMessage("surface busy");break;
	case DD_OK : oneMessage("drawn ok");break;
	default : oneMessage("error");break;
	}
}

void showMessage(char* mess)
{
	// Displays a text message in a dialog box
	Sprite::getDD()->FlipToGDISurface();
	if (mess==NULL) MessageBox(ghwnd,"NULL",TITLE,MB_OK);
	else MessageBox( ghwnd, mess, TITLE, MB_OK );
}

void showMessage(char* mess,int num)
{
	// Displays a text message followed by an integer
	Sprite::getDD()->FlipToGDISurface();
	char s[200];
	strcpy(s,mess);
	sprintf(s+strlen(mess),"%d",num);
	MessageBox( ghwnd, s, TITLE, MB_OK );
}

void showMessage(char* m1 , char* m2)
{
	// Merges two string together and displays result
	char* mess=new char[strlen(m1)+strlen(m2)+1];
	memcpy(mess,m1,strlen(m1));
	memcpy(mess+strlen(m1),m2,strlen(m2));
	*(mess+strlen(m1)+strlen(m2))=0;
	MessageBox(ghwnd , mess , TITLE , MB_OK);
}

void oneMessage(char* mess)
{
	// Displays a text message. After this function is called 10 times,
	// further calls are ignored (used for debugging)
	static int first=0;
	if (first<10) {	showMessage(mess); first++;}
}

void oneMessage(char* mess,int num)
{
	static int first=0;
	if (first<10) {	showMessage(mess,num); first++;}
}

void oneMessage(char* m1, char* m2)
{
	static int first=0;
	if (first<10) { showMessage(m1,m2); first++; }
}

BOOL OpenSaveDialog(char* buffer)
{
	// opens a standard Window95 save dialog box and returns the selected filename
	char tb[260];
	OPENFILENAME* fo = new OPENFILENAME;
	memset(fo,'\0',sizeof(OPENFILENAME));
	fo->lStructSize = sizeof(OPENFILENAME);
	fo->hwndOwner = ghwnd;
	fo->lpstrFilter = "Chuckie Save Files\0*.egg\0\0";
	fo->lpstrFile = tb; tb[0]='\0';
	fo->nMaxFile = 260;
	fo->Flags = OFN_FILEMUSTEXIST + OFN_LONGNAMES + OFN_HIDEREADONLY;
	Sprite::getDD()->FlipToGDISurface();
	BOOL res = GetSaveFileName(fo);
	strcpy(buffer,tb);
	return res;
}

BOOL OpenLoadDialog(char* buffer)
{
	// Presents a standard Window95 load dialog and returns the selected filename
	char tb[260];
	OPENFILENAME* fo = new OPENFILENAME;
	memset(fo,'\0',sizeof(OPENFILENAME));
	fo->lStructSize = sizeof(OPENFILENAME);
	fo->hwndOwner = ghwnd;
	fo->lpstrFilter = "Chuckie Save Files\0*.egg\0\0";
	fo->lpstrFile = tb; tb[0]='\0';
	fo->nMaxFile = 260;
	fo->Flags = OFN_FILEMUSTEXIST + OFN_LONGNAMES + OFN_HIDEREADONLY;
	Sprite::getDD()->FlipToGDISurface();
	BOOL res = GetOpenFileName(fo);
	strcpy(buffer,tb);
	return res;
}

void sizeMe(BOOL fs, int xs, int ys)
{
	Sprite::setNoRestore(TRUE);
	if (!fs)
	{
		Sprite::loseSurfaces();

//		SetWindowLong(ghwnd, GWL_EXSTYLE, WS_OVERLAPPEDWINDOW);

		RECT rc;
		SetRect(&rc,0,0,xs,ys);
		AdjustWindowRectEx(&rc, GetWindowLong(ghwnd, GWL_STYLE),
		GetMenu(ghwnd) != NULL, GetWindowLong(ghwnd, GWL_EXSTYLE));

		int left = (GetSystemMetrics(SM_CXSCREEN)-rc.right+rc.left)/2;
		int top = (GetSystemMetrics(SM_CYSCREEN)-32-rc.bottom+rc.top)/2;

		SetWindowPos(ghwnd, NULL, left,
			top, rc.right-rc.left, rc.bottom-rc.top,
			SWP_NOZORDER|SWP_NOACTIVATE);

	}
	else
	{
		Sprite::loseSurfaces();
//		SetWindowLong(ghwnd, GWL_EXSTYLE, WS_POPUP);
	}
	Sprite::setSize(fs,xs,ys);
	game.setSize(fs,xs,ys);
	scrnWidth=xs; scrnHeight=ys;
	fullScreen=fs;
//	ShowCursor(!fullScreen);
	ShowCursor(FALSE);
	SetCursor(LoadCursor(0,IDC_ARROW));
	Sprite::setNoRestore(FALSE);
}

void setFps(int v, BOOL display)
{
	if (v>1000) v=1000;
	if (v<1) v=1;
	fps=v;
	if (display) game.sendMessage(Game::SETFPS,v);
	MENUITEMINFO menuiteminfo;
	memset(&menuiteminfo,'\0',sizeof(menuiteminfo));
	menuiteminfo.cbSize=sizeof(menuiteminfo);
	menuiteminfo.fMask=MIIM_TYPE;
	menuiteminfo.fType=MFT_STRING;
	menuiteminfo.dwTypeData=new char[11];
	sprintf(menuiteminfo.dwTypeData, "%d", fps);

//	itoa(fps,menuiteminfo.dwTypeData,10);
//	menuiteminfo.cch=strlen((char*)menuiteminfo.dwTypeData);
	menuiteminfo.wID=ID_OPTIONS_FPS;
	SetMenuItemInfo(GetMenu(ghwnd),ID_OPTIONS_FPS,FALSE,&menuiteminfo);
	delete[] menuiteminfo.dwTypeData;
}

void tickLevelFile(int num)
{
	for(int c=0; c<noOfLevelFiles; c++)
		CheckMenuItem(GetMenu(ghwnd),ID_FILE_LEVEL+c,MF_UNCHECKED);
	CheckMenuItem(GetMenu(ghwnd),ID_FILE_LEVEL+num,MF_CHECKED);
}

void updateFrame( void )
{
	BOOL res = game.iterate();

	if (res)
		if (game.draw())
			Sprite::flipScreens();

	if (Sprite::getSurfacesLost())
	{
		game.restore();
		Sprite::setSurfacesLost(FALSE);
	}
}

static void finiObjects( void )
{

	for(int c=0; c<noOfLevelFiles; c++)
		delete levelFiles[c];

	game.closeDown();
	Sound::closeDown();
	Sprite::closeDown();
}

static void OnMouseMove(void*,int x,int y,UINT)
{
	game.mouseMove(x,y);
}

static void OnMouseLeftDown(void*,BOOL,int x,int y,UINT)
{
	game.mouseLeftDown(x,y);
}

static void OnMouseLeftUp(void*,int x,int y,UINT)
{
	game.mouseLeftUp(x,y);
}

static void OnMouseRightDown(void*,BOOL,int x,int y,UINT)
{	game.mouseRightDown(x,y);	}

static void OnMouseRightUp(void*,int x,int y,UINT)
{	game.mouseRightUp(x,y);	}

long FAR PASCAL WindowProc( HWND hWnd, UINT message,
							WPARAM wParam, LPARAM lParam )
{
	static int cc = 0;
	switch( message )
	{
	case WM_ACTIVATEAPP:
		bActive = wParam;
		break;

	case WM_SETCURSOR:
		ShowCursor(!fullScreen);
		return TRUE;

//	case WM_NCMOUSEMOVE:
//		showMessage("ncmousemove");
//		ShowCursor(TRUE);
//		return TRUE;

	case WM_CREATE:
		break;

	case WM_KEYDOWN:
		switch( wParam )
		{

		case VK_ESCAPE:
			if (game.getState()==Game::MENU)
			{
				if (fullScreen)
				{
					sizeMe(FALSE,windowWidth,windowHeight);
					static char jj[2];
					jj[0]=windowWidth/320; jj[1]='\0';
					WritePrivateProfileString("Chuckie","ScreenMode",jj,iniFile);
				}
				else
					PostMessage(hWnd, WM_CLOSE, 0, 0);
			}
			if (game.getState()!=Game::MENU)
				game.keyDown((char)1);
			break;

		case VK_TAB:
			if (game.save("boss.egg")==TRUE)
				WritePrivateProfileString("Chuckie","LoadBoss","1",iniFile);
			PostMessage(hWnd, WM_CLOSE, 0, 0);
			break;

		case VK_F12:
			PostMessage(hWnd, WM_CLOSE, 0, 0);
			break;

		default:
			game.keyDown((char)wParam);
		}
		break;

	case WM_KEYUP:
		game.keyUp((char)wParam);
		break;

	HANDLE_MSG(NULL, WM_MOUSEMOVE, OnMouseMove);
	HANDLE_MSG(NULL, WM_LBUTTONDOWN, OnMouseLeftDown);
	HANDLE_MSG(NULL, WM_LBUTTONUP, OnMouseLeftUp);
	HANDLE_MSG(NULL, WM_RBUTTONDOWN, OnMouseRightDown);
	HANDLE_MSG(NULL, WM_RBUTTONUP, OnMouseRightUp);

	case WM_DESTROY:
		finiObjects();
		PostQuitMessage( 0 );
		break;

	case WM_SYSCOMMAND:
		if((wParam&0xFFF0)==SC_SCREENSAVE || (wParam&0xFFF0)==SC_MONITORPOWER)
			return 0;
		break;

	case WM_COMMAND:
		switch(LOWORD(wParam))
		{
			case ID_FILE_SNAPSHOT:
				keybd_event(VK_SNAPSHOT,0,NULL,NULL);
				break;
			case ID_FILE_SAVE:
				if (game.getState()==Game::PLAYING || game.getState()==Game::DONELEVEL
					|| game.getState()==Game::STARTINGGAME || game.getState()==Game::PLAYERDEATH
					|| game.getState()==Game::GAMEOVER || game.getState()==Game::PLAYERPROMPT)
				{
					char savBuf[300];
					if (OpenSaveDialog(savBuf))
						game.save(savBuf);
				}
				else
					showMessage("Can only save when a game is being played.");
				break;
			case ID_FILE_OPEN:
				if (game.getState()==Game::MENU || game.getState()==Game::DRAWINGMENU)
				{
					char loadBuf[300];
					if (OpenLoadDialog(loadBuf))
						game.load(loadBuf);
				}
				else
					showMessage("Cannot load while a game is being played.");
				break;
			case ID_FILE_EXIT:
				PostMessage(hWnd, WM_CLOSE, 0, 0);
				break;
			case ID_SIZE_FULLSCREEN:
				if (fullScreen) CheckMenuItem(GetMenu(ghwnd),ID_SIZE_FULLSCREEN,MF_UNCHECKED);
				else CheckMenuItem(GetMenu(ghwnd),ID_SIZE_320x256+scrnWidth/320-1,MF_UNCHECKED);
				if (fullScreen)
				{
					sizeMe(FALSE,windowWidth,windowHeight);
					static char j[2];
					j[0]=(windowWidth/320)+48; j[1]='\0';
					WritePrivateProfileString("Chuckie","ScreenMode",j,iniFile);
					CheckMenuItem(GetMenu(ghwnd),ID_SIZE_320x256+windowWidth/320-1,MF_CHECKED);
				}
				else
				{
					windowWidth=scrnWidth;
					windowHeight=scrnHeight;
					sizeMe(TRUE,800,600);
					WritePrivateProfileString("Chuckie","ScreenMode","0",iniFile);
					CheckMenuItem(GetMenu(ghwnd),ID_SIZE_FULLSCREEN,MF_CHECKED);
				}
				break;
			case ID_OPTIONS_PERSISTENTSCORES:
				options.persistentScores=!options.persistentScores;
				game.sendMessage(Game::PERSISTENTSCORES,options.persistentScores?1:0);
				CheckMenuItem(GetMenu(ghwnd),ID_OPTIONS_PERSISTENTSCORES,options.persistentScores?MF_CHECKED:MF_UNCHECKED);
				WritePrivateProfileString("Chuckie","PersistentScores",options.persistentScores?"1":"0",iniFile);
				break;
			case ID_OPTIONS_PERSISTENTKEYS:
				options.persistentKeys=!options.persistentKeys;
				game.sendMessage(Game::PERSISTENTKEYS,options.persistentKeys?1:0);
				CheckMenuItem(GetMenu(ghwnd),ID_OPTIONS_PERSISTENTKEYS,options.persistentKeys?MF_CHECKED:MF_UNCHECKED);
				WritePrivateProfileString("Chuckie","PersistentKeys",options.persistentKeys?"1":"0",iniFile);
				break;
			case ID_OPTIONS_EXTENDEDJUMPS:
				options.extendedJumps=!options.extendedJumps;
				game.sendMessage(Game::EXTENDEDJUMPS,options.extendedJumps?1:0);
				CheckMenuItem(GetMenu(ghwnd),ID_OPTIONS_EXTENDEDJUMPS,options.extendedJumps?MF_CHECKED:MF_UNCHECKED);
				break;
			case ID_OPTIONS_USEJOYSTICK:
				useJoystick=!useJoystick;
				CheckMenuItem(GetMenu(ghwnd),ID_OPTIONS_USEJOYSTICK,
					useJoystick?MF_CHECKED:MF_UNCHECKED);
				game.sendMessage(Game::USEJOYSTICK,(int)useJoystick);
				WritePrivateProfileString("Chuckie","UseJoystick",useJoystick?"1":"0",iniFile);
				break;
			case ID_OPTIONS_NOSOUND:
				CheckMenuItem(GetMenu(ghwnd),ID_OPTIONS_NOSOUND+soundMode,MF_UNCHECKED);
				soundMode=0;
				CheckMenuItem(GetMenu(ghwnd),ID_OPTIONS_NOSOUND,MF_CHECKED);
				game.sendMessage(Game::SOUNDMODE,0);
				WritePrivateProfileString("Chuckie","SoundMode","0",iniFile);
				break;
			case ID_OPTIONS_MONO:
				CheckMenuItem(GetMenu(ghwnd),ID_OPTIONS_NOSOUND+soundMode,MF_UNCHECKED);
				soundMode=1;
				CheckMenuItem(GetMenu(ghwnd),ID_OPTIONS_MONO,MF_CHECKED);
				game.sendMessage(Game::SOUNDMODE,1);
				WritePrivateProfileString("Chuckie","SoundMode","1",iniFile);
				break;
			case ID_OPTIONS_STEREO:
				CheckMenuItem(GetMenu(ghwnd),ID_OPTIONS_NOSOUND+soundMode,MF_UNCHECKED);
				soundMode=2;
				CheckMenuItem(GetMenu(ghwnd),ID_OPTIONS_STEREO,MF_CHECKED);
				game.sendMessage(Game::SOUNDMODE,2);
				WritePrivateProfileString("Chuckie","SoundMode","2",iniFile);
				break;
			case ID_OPTIONS_FULLSTEREO:
				CheckMenuItem(GetMenu(ghwnd),ID_OPTIONS_NOSOUND+soundMode,MF_UNCHECKED);
				soundMode=3;
				CheckMenuItem(GetMenu(ghwnd),ID_OPTIONS_FULLSTEREO,MF_CHECKED);
				game.sendMessage(Game::SOUNDMODE,3);
				WritePrivateProfileString("Chuckie","SoundMode","3",iniFile);
				break;
			case ID_OPTIONS_SPEEDUP:
				setFps(fps+1,TRUE);
				break;
			case ID_OPTIONS_SLOWDOWN:
				setFps(fps-1,TRUE);
				break;
			case ID_OPTIONS_STARTLEVEL:
				showMessage("This menu item has not been implemented. Edit chuckie.ini and reload instead.");
				break;
			case ID_OPTIONS_RESET:
				showMessage("To reset the default options close the program, delete the file chuckie.ini and reload.");
				break;
			case ID_HELP_ABOUT:
				std::stringstream str;
				str << "Chuckie Egg for Windows 95 v" 
					<< VERSION_STRING
					<< "\nby Mike Elson\nEmail: mike@elson.com\nHomepage: vroomfondel.net/chuckie";
				showMessage(const_cast<char*>(str.str().c_str()));
				break;
		}

		// dynamicaly built size menu

		if (LOWORD(wParam)>=ID_SIZE_320x256 && LOWORD(wParam)<ID_SIZE_320x256+maxMagnify)
		{
			int menu_item=ID_SIZE_320x256+(scrnWidth/320)-1;
			if (fullScreen) CheckMenuItem(GetMenu(ghwnd),ID_SIZE_FULLSCREEN,MF_UNCHECKED);
			else CheckMenuItem(GetMenu(ghwnd),menu_item,MF_UNCHECKED);
//			showMessage("menu_item = ",menu_item);
//			showMessage("windowwidth = ",windowWidth);
			sizeMe(FALSE,(LOWORD(wParam)+1-ID_SIZE_320x256)*320,(LOWORD(wParam)+1-ID_SIZE_320x256)*256);
			static char jj[2]; jj[0]=LOWORD(wParam)+49-ID_SIZE_320x256; jj[1]='\0';
			WritePrivateProfileString("Chuckie","ScreenMode",jj,iniFile);
			CheckMenuItem(GetMenu(ghwnd),LOWORD(wParam),MF_CHECKED);
		}

		// levels file menu

		if (LOWORD(wParam)>=ID_FILE_LEVEL && LOWORD(wParam)<=ID_FILE_LEVEL+noOfLevelFiles)
		{
			if (game.getState()==Game::MENU || game.getState()==Game::DRAWINGMENU)
			{
				game.loadLevels(levelFiles[LOWORD(wParam)-ID_FILE_LEVEL]);
				tickLevelFile(LOWORD(wParam)-ID_FILE_LEVEL);
			}
			else
				showMessage("Cannot load levels with a game in progress.");
		}
	}

	return DefWindowProc(hWnd, message, wParam, lParam);

} /* WindowProc */

/*
 * This function is called if the initialization function fails
 */
BOOL initFail( HWND h )
{
	finiObjects();
	MessageBox( h, "DirectX Init FAILED", TITLE, MB_OK );
	DestroyWindow( h );
	return FALSE;

} /* initFail */

/*
 * doInit - do work required for every instance of the application:
 *                create the window, initialize data
 */
static BOOL doInit( HINSTANCE hInstance, int nCmdShow )
{
	HWND                hwnd;
	WNDCLASS            wc;
	memset(&wc, 0, sizeof(wc));

	// read initial screen size from chuckie.ini
	GetModuleFileName(hInstance,exeFile,256);
	GetModuleFileName(hInstance,appDir,256);
	char* ptr=appDir+strlen(appDir)-1;
	while(*ptr--!='\\'); ptr[2]='\0';
	strcpy(iniFile,appDir); strcat(iniFile,"Chuckie.ini");

	scrnWidth=GetSystemMetrics(SM_CXSCREEN);
	scrnHeight=GetSystemMetrics(SM_CYSCREEN);
	maxMagnify=(int)(scrnWidth/320);
	if ((int)(scrnHeight/256) < maxMagnify) maxMagnify=(int)(scrnHeight/256);
	if (maxMagnify>9) maxMagnify=9;

	int s = GetPrivateProfileInt("Chuckie","ScreenMode",1,iniFile);
	if (s>maxMagnify) s=maxMagnify;

	fullScreen=(s==0)?TRUE:FALSE;
	scrnWidth=(s==0)?800:s*320;
	scrnHeight=(s==0)?600:s*256;
	windowWidth=(s==0)?320:s*320;
	windowHeight=(s==0)?256:s*256;

	/*
	 * set up and register window class
	 */
	wc.style = CS_HREDRAW | CS_VREDRAW;
	wc.lpfnWndProc = WindowProc;
	wc.cbClsExtra = 0;
	wc.cbWndExtra = 0;
	wc.hInstance = hInstance;
	wc.hIcon = LoadIcon( hInstance, "Chuckie");
	wc.hCursor = LoadCursor( NULL, IDC_ARROW );
	wc.hbrBackground = GetStockBrush(BLACK_BRUSH);
	wc.lpszMenuName = MAKEINTRESOURCE(IDR_MENU);
	wc.lpszClassName = NAME;
	RegisterClass( &wc );

	hCursor=wc.hCursor;
	hAccel = LoadAccelerators(hInstance, "ChuckieAcc");
	/*
	 * create a window
	 */

	hwnd = CreateWindowEx(
		0,
		NAME,
		TITLE,
		WS_OVERLAPPEDWINDOW,
		(GetSystemMetrics(SM_CXSCREEN)-scrnWidth-7)/2,
		max((GetSystemMetrics(SM_CYSCREEN)-scrnHeight-77)/2,0),
		scrnWidth+7,//320,//scrnWidth,//GetSystemMetrics(SM_CXSCREEN),
		scrnHeight+45,//256,//scrnHeight,//GetSystemMetrics(SM_CYSCREEN),
		NULL,
		NULL,
		hInstance,
		NULL );

	if( !hwnd )
		return FALSE;

	ghwnd=hwnd;
	ghinstance=hInstance;

	ShowWindow( hwnd, nCmdShow );
	UpdateWindow( hwnd );

	if (fullScreen) SetWindowLong(ghwnd, GWL_EXSTYLE, WS_POPUP);

    if (readProfileInt("HighColour")==1) Sprite::setHicolour(TRUE);
	if (Sprite::setup(hwnd,fullScreen,scrnWidth,scrnHeight)==FALSE)
		initFail(hwnd);

	if (GetPrivateProfileInt("Chuckie","SoundEnabled",1,iniFile)==1)
		Sound::setup(hwnd);
	else
		Sound::setup(NULL);

	if (game.init(fullScreen,scrnWidth,scrnHeight) == FALSE)
		initFail(hwnd);
	game.setIniFile(iniFile);

	options = game.getOptions();
	useJoystick=options.useJoystick;
	soundMode=options.soundMode;
	fps=options.framesPerSecond;
	setFps(fps,FALSE);

	// dynamic build of size menu
	MENUITEMINFO menuiteminfo;
	memset(&menuiteminfo,'\0',sizeof(menuiteminfo));
	menuiteminfo.cbSize=sizeof(menuiteminfo);
	// add separator first
	menuiteminfo.fMask=MIIM_TYPE|MIIM_ID;
	menuiteminfo.wID=ID_SIZE_SEPARATOR;
	menuiteminfo.fType=MFT_SEPARATOR;
	menuiteminfo.dwTypeData=NULL;
	InsertMenuItem(GetMenu(ghwnd),ID_SIZE_FULLSCREEN,FALSE,&menuiteminfo);
	char menutext[20];
	if (maxMagnify>1)
		for(int m=1; m<maxMagnify; m++)
		{
			menuiteminfo.wID=20000+m;
			menuiteminfo.fType=MFT_STRING;
			itoa((m+1)*320,menutext,10);
			strcat(menutext,"x");
			itoa((m+1)*256,menutext+strlen(menutext),10);
			menuiteminfo.dwTypeData=menutext;
			menuiteminfo.cch=strlen(menuiteminfo.dwTypeData);
			InsertMenuItem(GetMenu(ghwnd),ID_SIZE_SEPARATOR,FALSE,&menuiteminfo);
		}

	if (useJoystick)
		CheckMenuItem(GetMenu(ghwnd),ID_OPTIONS_USEJOYSTICK,MF_CHECKED);

	CheckMenuItem(GetMenu(ghwnd),ID_OPTIONS_NOSOUND+soundMode,MF_CHECKED);

	if (fullScreen) CheckMenuItem(GetMenu(ghwnd),ID_SIZE_FULLSCREEN,MF_CHECKED);
	else CheckMenuItem(GetMenu(ghwnd),ID_SIZE_320x256+windowWidth/320-1,MF_CHECKED);

	if (options.persistentScores)
		CheckMenuItem(GetMenu(ghwnd),ID_OPTIONS_PERSISTENTSCORES,MF_CHECKED);
	if (options.persistentKeys)
		CheckMenuItem(GetMenu(ghwnd),ID_OPTIONS_PERSISTENTKEYS,MF_CHECKED);
	if (options.extendedJumps)
		CheckMenuItem(GetMenu(ghwnd),ID_OPTIONS_EXTENDEDJUMPS,MF_CHECKED);


	// display start level in menu

	memset(&menuiteminfo,'\0',sizeof(menuiteminfo));
	menuiteminfo.cbSize=sizeof(menuiteminfo);
	menuiteminfo.fMask=MIIM_TYPE;
	menuiteminfo.fType=MFT_STRING;
	menuiteminfo.dwTypeData=new char[5];
	itoa(options.levelJump+1,menuiteminfo.dwTypeData,10);
//	menuiteminfo.cch=strlen((char*)menuiteminfo.dwTypeData);
	menuiteminfo.wID=ID_OPTIONS_STARTLEVEL;
	SetMenuItemInfo(GetMenu(ghwnd),ID_OPTIONS_STARTLEVEL,FALSE,&menuiteminfo);
	delete[] menuiteminfo.dwTypeData;

	// look for text file containing level data, and build level selection menu

	char startLevel[100]; strcpy(startLevel,readProfileStr("LevelFile"));
	if (!strcmp(startLevel,"levels.txt")) tickLevelFile(0);
	levelFiles[0] = new char[11];
	strcpy(levelFiles[0],"levels.txt"); levelFiles[0][10]='\0';

	//ffblk fileStruct;
	_finddata_t fileStruct;

//struct _finddata_t {
//    unsigned    attrib;
//    time_t      time_create;    /* -1 for FAT file systems */
//    time_t      time_access;    /* -1 for FAT file systems */
//    time_t      time_write;
//    _fsize_t    size;
//    char        name[260];
//};


	char searchDir[256];
	strcpy(searchDir, appDir); strcat(searchDir, "\\*.txt");
	int ff_handle=_findfirst(searchDir, &fileStruct);
	if (ff_handle==-1)
	{
		showMessage("Fatal Error - Levels file cannot be loaded.");
		initFail(hwnd);
	}
	else
	{
		int idCode = ID_FILE_LEVEL + 1;
		noOfLevelFiles = 1;
		int res = 0;
		while(res!=-1)
		{
			int handle=open(fileStruct.name,O_RDONLY);
			if (handle!=-1 && strcmp(fileStruct.name,"levels.txt"))
			{
				read(handle,searchDir,8);
				if (!memcmp(searchDir,"[levels]",8))
				{
					levelFiles[noOfLevelFiles] = new char[strlen(fileStruct.name)+1];
					strcpy(levelFiles[noOfLevelFiles],fileStruct.name);
					memset(&menuiteminfo,'\0',sizeof(menuiteminfo));
					menuiteminfo.cbSize=sizeof(menuiteminfo);
					menuiteminfo.fMask=MIIM_TYPE|MIIM_ID;
					menuiteminfo.fType=MFT_STRING;
					menuiteminfo.wID=idCode++;
					menuiteminfo.fType=MFT_STRING;
					menuiteminfo.dwTypeData=fileStruct.name;
					menuiteminfo.cch=strlen(menuiteminfo.dwTypeData);
					InsertMenuItem(GetMenu(ghwnd),ID_FILE_LEVEL,FALSE,&menuiteminfo);
					if (!strcmp(startLevel,fileStruct.name))
						tickLevelFile(idCode-ID_FILE_LEVEL-1);
					noOfLevelFiles++;
				}
			}
			res=_findnext(ff_handle, &fileStruct);
		}
	}

	// check if data file has been dblclicked on, or boss key used previous game
	if (strlen(commandLine)>0)
		game.load(commandLine);
	else
	{
		if (readProfileInt("LoadBoss",0)==1)
			game.load("boss.egg");
	}
	WritePrivateProfileString("Chuckie","LoadBoss","0",iniFile);

	// patch registry to give .egg files nice icon and associate file
	// with chuckie.exe

	HKEY hkey;
	unsigned long exists;
	if ( RegCreateKeyEx(HKEY_CLASSES_ROOT, ".egg", 0, 0, NULL, KEY_ALL_ACCESS,
		NULL, &hkey, &exists) != ERROR_SUCCESS)
		showMessage("RegCreateKey (.egg) failed");
	else
	{
		if (RegSetValueEx(hkey, NULL, 0, REG_SZ, reinterpret_cast<const unsigned char*>("ChuckieSaveFile"), 16) != ERROR_SUCCESS)
			showMessage("RegSetValueEx (.egg) failed");
/*		else
		{
			if (exists==REG_CREATED_NEW_KEY)
				showMessage("New");
			else
				showMessage("Existing");
		}*/
	}

	if ( RegCreateKeyEx(HKEY_CLASSES_ROOT, "ChuckieSaveFile", 0, 0, NULL, KEY_ALL_ACCESS,
		NULL, &hkey, &exists) != ERROR_SUCCESS)
		showMessage("RegCreateKey (ChuckieSaveFile) failed");
	else
	{
		if (RegSetValueEx(hkey, NULL, 0, REG_SZ, reinterpret_cast<unsigned char*>("Chuckie Egg saved game"), 23) != ERROR_SUCCESS)
			showMessage("RegSetValueEx (ChuckieSaveFile) failed");
		else
		{
			HKEY iconKey;
			if ( RegCreateKeyEx(hkey, "DefaultIcon", 0, 0, NULL, KEY_ALL_ACCESS,
				NULL, &iconKey, &exists) != ERROR_SUCCESS)
				showMessage("RegCreateKey (DefaultIcon) failed");
			else
			{
				char iconFile[256]; strcpy(iconFile,exeFile); strcat(iconFile,",1");
				if (RegSetValueEx(iconKey, NULL, 0, REG_SZ, reinterpret_cast<unsigned char*>(iconFile), strlen(iconFile)+1) != ERROR_SUCCESS)
					showMessage("RegSetValueEx (DefaultIcon) failed");
				else
				{
					if ( RegCreateKeyEx(hkey, "Shell", 0, 0, NULL, KEY_ALL_ACCESS,
						NULL, &hkey, &exists) != ERROR_SUCCESS)
						showMessage("RegCreateKey (Shell) failed");
					else
					{
						if (RegSetValueEx(hkey, NULL, 0, REG_SZ, reinterpret_cast<const unsigned char*>("Run"), 4) != ERROR_SUCCESS)
							showMessage("RegSetValueEx (Run) failed");
						else
						{
							if ( RegCreateKeyEx(hkey, "Run", 0, 0, NULL, KEY_ALL_ACCESS,
								NULL, &hkey, &exists) != ERROR_SUCCESS)
								showMessage("RegCreateKey (Run) failed");
							else
							{
								if (RegSetValueEx(hkey, NULL, 0, REG_SZ, reinterpret_cast<const unsigned char*>("&Run"), 5) != ERROR_SUCCESS)
									showMessage("RegSetValueEx (Run) failed");
								else
								{
									if ( RegCreateKeyEx(hkey, "command", 0, 0, NULL, KEY_ALL_ACCESS,
										NULL, &hkey, &exists) != ERROR_SUCCESS)
										showMessage("RegCreateKey (Command) failed");
									else
									{
										char com[256]; strcpy(com,exeFile); strcat(com," %1");
										if (RegSetValueEx(hkey, NULL, 0, REG_SZ, reinterpret_cast<unsigned char*>(com), strlen(com)) != ERROR_SUCCESS)
											showMessage("RegSetValueEx (%1) failed");
									}
								}
							}
						}
					}
				}
			}
		}
	}

	SetCursor(LoadCursor(0,IDC_ARROW));
//	ShowCursor(FALSE);

	return TRUE;
} /* doInit */

/*
 * WinMain - initialization, message loop
 */
//int PASCAL WinMain( HINSTANCE hInstance, HINSTANCE hPrevInstance,
//						LPSTR lpCmdLine, int nCmdShow)
int PASCAL WinMain( HINSTANCE hInstance, HINSTANCE ,
						LPSTR lpCmdLine, int nCmdShow)
{
	MSG         msg;
	strcpy(commandLine, lpCmdLine);

	if( !doInit( hInstance, nCmdShow ) )
	{
		return FALSE;
	}

	while( 1 )
	{
		if( PeekMessage( &msg, NULL, 0, 0, PM_NOREMOVE ) )
		{
			if( !GetMessage( &msg, NULL, 0, 0 ) )
				return msg.wParam;
			if (!TranslateAccelerator(ghwnd, hAccel, &msg))
			{
				TranslateMessage(&msg);
				DispatchMessage(&msg);
			}
		}
		else if( bActive )
		{
			updateFrame();
		}
		else
		{
			// make sure we go to sleep if we have nothing else to do
			WaitMessage();
		}
	}
} /* WinMain */

int readProfileInt(char* str)
{	return readProfileInt(str,0);	}

int readProfileInt(char* str, int def)
{
	return GetPrivateProfileInt("Chuckie",str,def,iniFile);
}

char* readProfileStr(char* str)
{
	static char res[300];
	if (GetPrivateProfileString("Chuckie",str,"NULL",res,200,iniFile))
		return res;
	else
		return NULL;
}

void writeProfileInt(char* str, int num)
{
	char buf[30];
	WritePrivateProfileString("Chuckie",str,itoa(num,buf,10),iniFile);
}

void writeProfileStr(char* str, char* dat)
{
	WritePrivateProfileString("Chuckie",str,dat,iniFile);
}



