// sprite.cpp

// Contains low level calls for creating sprites, loading a .bmp file into
// a sprite and plotting to screen.
// This file also contains routines to rescale and rotate sprites.

// Many section of this file, in particualr the 'Load' routine and basic
// sprite drawing are based on sample code and tutorials from the DirectX SDK by
// Microsoft.
// Used with the permission of Microsoft.

#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <windowsx.h>
#include <ddraw.h>
#define DoneDDraw
#define DoneWindows

#include <fcntl.h>
#include <io.h>
#include <math.h>
const double M_PI = 3.141592;
#include <string.h>
//#include <mem.h>

#include "sprite.h"
#include "main.h"
#include "ddutil.h"
#include "LinkList.h"

LinkedList<Sprite*> Sprite::list;
LPDIRECTDRAW Sprite::lpDD;
int Sprite::scrnWidth;
int Sprite::scrnHeight;
int Sprite::count=0;
LPDIRECTDRAWSURFACE Sprite::lpDDSBack;
LPDIRECTDRAWSURFACE Sprite::lpDDSPrimary;
LPDDCOLORKEY Sprite::lpDDColorKey;
LPDIRECTDRAWPALETTE Sprite::pDDpal;
unsigned char* Sprite::pFont;
BOOL Sprite::surfacesLost;
BOOL Sprite::fullScreen;
LPDIRECTDRAWCLIPPER Sprite::lpDDClipper;
HWND Sprite::hwnd;
BOOL Sprite::noRestore;
BOOL Sprite::hicolour=FALSE;

BOOL Sprite::setup(void* hw,BOOL fs,int w,int h)
{
	// Start up DirectDraw

	Sprite::noRestore=FALSE;
	Sprite::hwnd = (HWND)hw;
	HRESULT ddrval;
	Sprite::scrnWidth=w; Sprite::scrnHeight=h;
	Sprite::fullScreen=fs;
	Sprite::lpDDSPrimary=NULL;
	Sprite::lpDDSBack=NULL;
	/*
	 * create the main DirectDraw object
	 */
	ddrval = DirectDrawCreate( NULL, &Sprite::lpDD, NULL );
	if( ddrval != DD_OK )
		return FALSE;

	createSurfaces();

	Sprite::lpDDColorKey=new DDCOLORKEY;

	Sprite::scrnWidth=scrnWidth; Sprite::scrnHeight=scrnHeight;
	Sprite::lpDDColorKey->dwColorSpaceLowValue = 0;//RGBQUAD(0,0,0);
	Sprite::lpDDColorKey->dwColorSpaceHighValue = 0;//RGBQUAD(15,15,15);
	Sprite::pDDpal = DDLoadPalette(Sprite::lpDD, "palette");
	if (Sprite::pDDpal==NULL) showMessage("Error loading palette files");
	ddrval = Sprite::lpDDSPrimary->SetPalette(Sprite::pDDpal);

//	int handle = open("font.dat",O_RDONLY);
//	if (handle==-1) {showMessage("Error reading font.dat");return FALSE;}

/*		HRSRC hrsrc = FindResource( GetModuleHandle(NULL) , "fontdata" , "data");
		if (hrsrc==NULL)
		{
			showMessage("Fatal error: Cannot locate level data");
			return FALSE;
		}
		int size=SizeofResource(GetModuleHandle(NULL), hrsrc);
		HGLOBAL hglobal = LoadResource( NULL, hrsrc );
		if (hglobal==NULL)
		{
			showMessage("Fatal error: Cannot load level resource");
			return FALSE;
		}
		BYTE* data = (BYTE*)LockResource( hglobal );
		if (data==NULL)
		{
			showMessage("Fatal error: Cannot lock resource data");
			return FALSE;
		}*/


	Sprite::pFont=new unsigned char[129*256];
//	memcpy(Sprite::pFont, data, size);
//	read(handle,Sprite::pFont,filelength(handle));
//	close(handle);

	Sprite::surfacesLost=FALSE;

	return TRUE;
}

void Sprite::closeDown()
{
//	delete[] Sprite::pFont;
	list.reset();
	Sprite* item=list.getNext();
	while(item!=NULL)
	{
		delete item;
		item=list.getNext();
	}
	if( Sprite::pDDpal != NULL )
	{
		Sprite::pDDpal->Release();
		Sprite::pDDpal = NULL;
	}
	if( lpDD != NULL )
	{
		if (!Sprite::fullScreen && lpDDSBack)
		{
			lpDDSBack->Release();
			lpDDSBack=NULL;
		}
		if( lpDDSPrimary != NULL )
		{
			lpDDSPrimary->Release();
			lpDDSPrimary = NULL;
		}
//		lpDD->Release();
//		lpDD = NULL;
	}

//	HRESULT ddrval = Sprite::lpDDSPrimary->SetPalette(NULL);
//	if (ddrval != DD_OK) {showMessage("Error deleting palette");showError(ddrval);}
}

void Sprite::flipScreens()
{
	// Swap over displayed screen and back screen

	HRESULT ddrval;
	// Flip the surfaces
	if (fullScreen==TRUE)
		while( 1 )
		{
			ddrval = Sprite::lpDDSPrimary->Flip( NULL, 0 );
			if( ddrval == DD_OK || ddrval == DDERR_SURFACELOST) break;
			if( ddrval != DDERR_WASSTILLDRAWING ) break;
		}
	else
		while(1)
		{
			static RECT srcRect, destRect;
			static POINT pt;
			srcRect.left=0; srcRect.top=0;
			srcRect.right=Sprite::scrnWidth; srcRect.bottom=Sprite::scrnHeight;
			destRect.top=0; destRect.left=0;
			destRect.bottom=Sprite::scrnHeight; destRect.right=Sprite::scrnWidth;
            pt.x=pt.y=0;
			ClientToScreen(Sprite::hwnd, &pt);
            OffsetRect(&destRect, pt.x, pt.y);
			ddrval = Sprite::lpDDSPrimary->Blt(&destRect, Sprite::lpDDSBack,
				&srcRect, 0, NULL);
			if( ddrval == DD_OK || ddrval == DDERR_SURFACELOST) break;
			if( ddrval != DDERR_WASSTILLDRAWING ) break;
		}

	if (ddrval==DDERR_SURFACELOST) Sprite::restoreAll();
}

Sprite::Sprite()
{	pdds=NULL;
	filename=NULL;
	list.addItem(this);
	pitch=-1;
	bLocked=FALSE;
}

Sprite::~Sprite()
{
	list.deleteItem(this);
//	if (pdds!=NULL) { pdds->Release(); pdds=NULL;	}
//	if (filename!=NULL) delete[] filename;
}

BOOL Sprite::load(char* name)
{
	return load(name,1);
}

BOOL Sprite::load(char* name, int magnify) // load and magnify
{
	if (pdds!=NULL) { pdds->Release(); pdds=NULL; }

	if ((pdds=DDLoadBitmap(Sprite::lpDD,name,0,0,magnify)) == NULL)
	{
		showMessage("Error loading sprite ",name);
		return FALSE;
	}

	DDSURFACEDESC ddsd;
	HRESULT hResult;

	ZeroMemory(&ddsd,sizeof(ddsd));
	ddsd.dwSize=sizeof(ddsd);

	if ((hResult=pdds->GetSurfaceDesc(&ddsd)) != DD_OK)
	{
		showMessage("Error retrieving surface description");
		if (hResult==DDERR_INVALIDOBJECT) showMessage("Invalid Object");
		if (hResult==DDERR_INVALIDPARAMS) showMessage("Invalid Params");
		return FALSE;
	}

	width=ddsd.dwWidth;
	height=ddsd.dwHeight;

	pdds->SetColorKey(DDCKEY_SRCBLT,Sprite::lpDDColorKey);
	mask=TRUE;

	filename=new char[strlen(name)+1];
	strcpy(filename,name);

	magnification=magnify;

	flipped=FALSE;

	return TRUE;
}

BOOL Sprite::load(char* name, int xsize, int ysize) // load and scale
{
	if (pdds!=NULL) { pdds->Release(); pdds=NULL; }

	if ((pdds=DDLoadScaledBitmap(Sprite::lpDD,name,0,0,xsize,ysize)) == NULL)
	{
		showMessage("Error loading sprite ",name);
		return FALSE;
	}

	DDSURFACEDESC ddsd;
	HRESULT hResult;

	ZeroMemory(&ddsd,sizeof(ddsd));
	ddsd.dwSize=sizeof(ddsd);

	if ((hResult=pdds->GetSurfaceDesc(&ddsd)) != DD_OK)
	{
		showMessage("Error retrieving surface description");
		if (hResult==DDERR_INVALIDOBJECT) showMessage("Invalid Object");
		if (hResult==DDERR_INVALIDPARAMS) showMessage("Invalid Params");
		return FALSE;
	}

	width=ddsd.dwWidth;
	height=ddsd.dwHeight;

	pdds->SetColorKey(DDCKEY_SRCBLT,Sprite::lpDDColorKey);
	mask=TRUE;

	filename=new char[strlen(name)+1];
	strcpy(filename,name);

	magnification=1;

	flipped=FALSE;

	return TRUE;
}

void Sprite::setSize(int xs,int ys)
{
	// Destroy any existing sprite data and set this instance to size a xs*ys
	// black image

	if (pdds!=NULL) { pdds->Release(); pdds=NULL; }
	DDSURFACEDESC ddsd;
	ZeroMemory(&ddsd, sizeof(ddsd));
	ddsd.dwSize = sizeof(ddsd);
	ddsd.dwFlags = DDSD_CAPS | DDSD_HEIGHT |DDSD_WIDTH;
	ddsd.ddsCaps.dwCaps = DDSCAPS_OFFSCREENPLAIN;// | DDSCAPS_SYSTEMMEMORY;
	ddsd.dwWidth = xs;
	ddsd.dwHeight = ys;
	if (Sprite::lpDD->CreateSurface(&ddsd, &pdds, NULL) != DD_OK)
		showMessage("Sprite::setSize: Error creating DD surface ",xs);
	width=xs; height=ys;
	mask=FALSE;
	flipped=FALSE;

	blacken();
}

void Sprite::blacken()
{
	DDBLTFX ddbltfx;
	ddbltfx.dwSize = sizeof(ddbltfx);
	ddbltfx.dwFillColor =0;
	if (pdds->Blt(	NULL,        // Destination
		NULL, NULL,  // Source rectangle
		DDBLT_COLORFILL, &ddbltfx)!=DD_OK) showMessage("error flood filling");
}

BOOL Sprite::draw(int x, int y)
{
	// Plot sprite to scale 1:1 at specified position.
	// Includes checks that x,y is on screen.
	// If partially off screen only a portion of the image is drawn

	if (x<=-width+1 || x>=Sprite::scrnWidth || y<=-height+1 || y>=Sprite::scrnHeight) return TRUE;

	static HRESULT ddrval;
	static RECT src,dest;
	src.left=0;
	src.right=width;
	src.top=0;
	src.bottom=height;
	dest.left=x;
	dest.right=x+width;
	dest.top=y;
	dest.bottom=y+height;

	// check for partial off-screen
	if (x<0)
		{ src.left=0-x; dest.left=0; dest.right=width+x; }
	if (y<0)
		{ src.top=0-y; dest.top=0; dest.bottom=height+y; }
	if (x+width>scrnWidth)
		{ src.right=width-(x+width-scrnWidth); dest.right=scrnWidth; }
	if (y+height>scrnHeight)
		{ src.bottom=height-(y+height-scrnHeight); dest.bottom=scrnHeight;}
	while( 1 )
	{
	   ddrval = Sprite::lpDDSBack->Blt(&dest, pdds,
							 &src,(mask==TRUE ? DDBLT_KEYSRC : FALSE),NULL);

		if (ddrval!=DDERR_WASSTILLDRAWING) break;
	}

	if (ddrval==DD_OK) return TRUE;
	if (ddrval == DDERR_SURFACELOST)
	{
		Sprite::restoreAll();
		return TRUE;
	}
	showError(ddrval);
	return FALSE;
}

BOOL Sprite::drawFlip(int x, int y)
{
	if (x<=-width+1 || x>=Sprite::scrnWidth || y<=-height+1 || y>=Sprite::scrnHeight) return TRUE;

	static HRESULT ddrval;
	static RECT src,dest;
	src.left=0;
	src.right=width;
	src.top=0;
	src.bottom=height;
	dest.left=x;
	dest.right=x+width;
	dest.top=y;
	dest.bottom=y+height;

	// check for partial off-screen
	if (x<0)
		{ src.left=0-x; dest.left=0; dest.right=width+x; }
	if (y<0)
		{ src.top=0-y; dest.top=0; dest.bottom=height+y; }
	if (x+width>scrnWidth)
		{ src.right=width-(x+width-scrnWidth); dest.right=scrnWidth; }
	if (y+height>scrnHeight)
		{ src.bottom=height-(y+height-scrnHeight); dest.bottom=scrnHeight;}

	static DDBLTFX bltfx;
	ZeroMemory(&bltfx,sizeof(bltfx));
	bltfx.dwSize=sizeof(bltfx);
	bltfx.dwDDFX=DDBLTFX_MIRRORLEFTRIGHT;
    
	while( 1 )
	{
	   ddrval = Sprite::lpDDSBack->Blt(&dest, pdds,
							 &src,(mask==TRUE ? DDBLT_KEYSRC|DDBLT_DDFX : DDBLT_DDFX),&bltfx);

		if (ddrval!=DDERR_WASSTILLDRAWING) break;
	}

	if (ddrval==DD_OK) return TRUE;
	if (ddrval == DDERR_SURFACELOST)
	{
		Sprite::restoreAll();
		return TRUE;
	}
	showError(ddrval);
	return FALSE;
}

BOOL Sprite::draw(int x, int y, float size)
{
	// Draws this instance at size (size : 1)

	static int xs,ys;
	xs=(int)((float)width*size);
	ys=(int)((float)height*size);

	// return if image too small
	if (xs<=1 || ys<=1) return TRUE;

	// return is fully off screen
	if (x<=-xs+1 || x+1>=Sprite::scrnWidth || y<=-ys+1 || y+1>=Sprite::scrnHeight) return TRUE;

	static HRESULT ddrval;
	static RECT src,dest;
	src.left=0;
	src.right=width;
	src.top=0;
	src.bottom=height;
	dest.left=x;
	dest.right=x+xs;
	dest.top=y;
	dest.bottom=y+ys;

	// if partially off screen calculate section to be drawm
	if (x<0)
		{ src.left=0-static_cast<int>(x/size); dest.left=0; dest.right=xs+x; }
	if (y<0)
		{ src.top=0-static_cast<int>(y/size); dest.top=0; dest.bottom=ys+y; }
	if (x+xs>scrnWidth)
		{ src.right=width-static_cast<int>((x+xs-scrnWidth)/size); dest.right=scrnWidth; }
	if (y+ys>scrnHeight)
		{ src.bottom=height-static_cast<int>((y+ys-scrnHeight)/size); dest.bottom=scrnHeight;}

	// prevent error message if sprite is at bottom of screen
	if (src.top==0 && src.bottom==0) src.bottom=1;

	while( 1 )
	{
		ddrval = Sprite::lpDDSBack->Blt(&dest, pdds,
							 &src,(mask==TRUE ? DDBLT_KEYSRC : FALSE),NULL);

		if (ddrval!=DDERR_WASSTILLDRAWING) break;
	}

	if (ddrval==DD_OK) return TRUE;

	if (ddrval==DDERR_INVALIDRECT)
	{
		oneMessage("source left = ",src.left);
		oneMessage("dest left = ",dest.left);
		oneMessage("source right = ",src.right);
		oneMessage("dest right = ",dest.right);
		oneMessage("source top = ",src.top);
		oneMessage("dest top = ",dest.top);
		oneMessage("source bottom = ",src.bottom);
		oneMessage("dest bottom = ",dest.bottom);
		return FALSE;
	}

	if (ddrval==DDERR_SURFACELOST)
	{
		Sprite::restoreAll();
		return TRUE;
	}

	showError(ddrval);
	return TRUE;
}

BOOL Sprite::draw(Sprite* target, int x, int y)
{
	static HRESULT ddrval;
	static RECT src,dest;
	src.left=0;
	src.right=width;
	src.top=0;
	src.bottom=height;
	dest.left=x;
	dest.right=x+width;
	dest.top=y;
	dest.bottom=y+height;
	while( 1 )
	{
		ddrval = target->pdds->Blt(&dest, pdds,
							 &src,(mask==TRUE ? DDBLT_KEYSRC : FALSE),NULL);

		if (ddrval!=DDERR_WASSTILLDRAWING) break;
	}

	if (ddrval==DD_OK) return TRUE;

	if (ddrval==DDERR_SURFACELOST)
	{
		Sprite::restoreAll();
		return TRUE;
	}

	showError(ddrval);
	return FALSE;
}

BOOL Sprite::draw(Sprite* target, int x, int y, RECT& src)
{
	static RECT dest;
	dest.left=x;
	dest.right=x+src.right-src.left;
	dest.top=y;
	dest.bottom=y+src.bottom-src.top;
	HRESULT ddrval;
	while( 1 )
	{
	   ddrval = target->pdds->Blt(&dest, pdds,
							 &src,(mask==TRUE ? DDBLT_KEYSRC : FALSE),NULL);

		if (ddrval!=DDERR_WASSTILLDRAWING) break;
	}

	if (ddrval==DD_OK) return TRUE;
	if (ddrval == DDERR_SURFACELOST)
	{
		Sprite::restoreAll();
		return TRUE;
	}
	showError(ddrval);
	return FALSE;
}

BOOL Sprite::draw(int x, int y, RECT& src)
{
	if (x<=(src.left-src.right) || x>=scrnWidth
		|| y<=(src.top-src.bottom) || y>=scrnHeight)
		return TRUE;

	static RECT dest;
	dest.left=x;
	dest.right=x+src.right-src.left;
	dest.top=y;
	dest.bottom=y+src.bottom-src.top;

	if (x<0) {src.left -=x; dest.left=0;}
	if (y<0) {src.top-=y; dest.top=0;}
	if ((y+src.bottom-src.top)>=scrnHeight) {src.bottom=src.top+scrnHeight-y;dest.bottom=scrnHeight;}
	if ((x+src.right-src.left)>=scrnWidth) {src.right=src.left+scrnWidth-x;dest.right=scrnWidth;}

	HRESULT ddrval;
	while( 1 )
	{
	   ddrval = Sprite::lpDDSBack->Blt(&dest, pdds,
							 &src,(mask==TRUE ? DDBLT_KEYSRC : FALSE),NULL);

		if (ddrval!=DDERR_WASSTILLDRAWING) break;
	}

	if (ddrval==DD_OK) return TRUE;
	if (ddrval == DDERR_SURFACELOST)
	{
		Sprite::restoreAll();
		return TRUE;
	}
	if (ddrval==DDERR_INVALIDRECT)
	{
		oneMessage("source left = ",src.left);
		oneMessage("dest left = ",dest.left);
		oneMessage("source right = ",src.right);
		oneMessage("dest right = ",dest.right);
		oneMessage("source top = ",src.top);
		oneMessage("dest top = ",dest.top);
		oneMessage("source bottom = ",src.bottom);
		oneMessage("dest bottom = ",dest.bottom);
		return FALSE;
	}
	showError(ddrval);
	return FALSE;
}

void Sprite::rotate(Sprite* source, int turn)
{
	// copies image of *source into self with rotation to turn
	// Units used are turn=64 := full 360 degree rotation

	DDSURFACEDESC desc;
	HRESULT result;

	setSize(source->width,source->height);
	desc.dwSize = sizeof( DDSURFACEDESC );
	do {
		result = (source->pdds)->Lock( NULL, &desc, DDLOCK_SURFACEMEMORYPTR, NULL );
		if( result != DD_OK && result != DDERR_WASSTILLDRAWING ) {showError(result);return;}
	} while( result == DDERR_WASSTILLDRAWING );
	if (result!=DD_OK) {showError(result);return;}
	BYTE* src=(BYTE*)desc.lpSurface;
	int srcPitch=desc.lPitch;
	do {
		result = pdds->Lock( NULL, &desc, DDLOCK_SURFACEMEMORYPTR, NULL );
		if( result != DD_OK && result != DDERR_WASSTILLDRAWING ) {showError(result);return;}
	} while( result == DDERR_WASSTILLDRAWING );
	if (result!=DD_OK) {showError(result);return;}
	BYTE* dest=(BYTE*)desc.lpSurface;
	int destPitch=desc.lPitch;

	for(int x=0; x<width; x++)
		for(int y=0; y<height; y++)
		{
			double angle;
			double fx,fy,fxs,fys;
			fx=static_cast<double>(x); fy=static_cast<double>(y); fxs=static_cast<double>(width); fys=static_cast<double>(height);
			if ((fx-static_cast<int>(fxs)/2)==0 && fy>static_cast<int>(fys/2)) angle=M_PI/2;
			if ((fx-static_cast<int>(fxs/2))==0 && fy<static_cast<int>(fys/2)) angle=3*M_PI/2;
			if (fy==fys/2 && fx>fxs/2) angle=0;
			if (fy==static_cast<int>(fys/2) && fx<static_cast<int>(fxs/2)) angle=M_PI;
			if (fx>static_cast<int>(fxs/2) && fy>static_cast<int>(fys/2)) angle=atan((fy-fys/2)/(fx-fxs/2));
			if (fx<fxs/2 && fy>fys/2) angle=M_PI-atan((fy-fys/2)/(fxs/2-fx));
			if (fx>fxs/2 && fy<fys/2) angle=M_PI*2-atan((fys/2-fy)/(fx-fxs/2));
			if (fx<fxs/2 && fy<fys/2) angle=M_PI+atan((fys/2-fy)/(fxs/2-fx));
			double dist=sqrt((fx-fxs/2)*(fx-fxs/2)+(fy-fys/2)*(fy-fys/2));
			angle+=M_PI/32*turn;
			double ftx=cos(angle)*dist+fxs/2;
			double fty=sin(angle)*dist+fys/2;

			int tx=(int)ftx; int ty=(int)fty;
			if (tx>=0 && tx<width && ty>=0 && ty<height)
				dest[x+y*destPitch]=src[tx+ty*srcPitch];
			else
				dest[x+y*destPitch]=0;
		}
	result=pdds->Unlock(NULL);
	if (result!=DD_OK) showError(result);
	result=(source->pdds)->Unlock(NULL);
	if (result!=DD_OK) showError(result);
	mask=source->mask;
	if (mask==TRUE) pdds->SetColorKey(DDCKEY_SRCBLT,Sprite::lpDDColorKey);
}

BYTE* Sprite::lock()
{
	// Before images can be directly modified we must 'lock' the surface.
	// This gives access to a pointer to the surface itself

	if (bLocked) return NULL;
	DDSURFACEDESC desc;
	HRESULT result;
	desc.dwSize = sizeof( DDSURFACEDESC );
	do {
		result = pdds->Lock( NULL, &desc, DDLOCK_SURFACEMEMORYPTR, NULL );
	} while( result == DDERR_WASSTILLDRAWING );
	if (result!=DD_OK) {showError(result);return NULL;}
	pitch=desc.lPitch;
	pSurface=(BYTE*)desc.lpSurface;
	bLocked=TRUE;
	return (BYTE*)desc.lpSurface;
}

void Sprite::unlock()
{
	// frees surface for normal drawing operations

	if (!bLocked) return;
	HRESULT result=pdds->Unlock(NULL);
	if (result!=DD_OK) showError(result);
	pitch=-1;
	bLocked=FALSE;
}

int Sprite::getPitch()
{	return pitch; }

int Sprite::getScrnWidth()
{	return Sprite::scrnWidth;	}

int Sprite::getScrnHeight()
{	return Sprite::scrnHeight;	}

void Sprite::drawText(char* s,int x,int y)
{
	// Basic text draw, outputs string at 's' to position x,y
	// Surface must be locked first

	if (!bLocked) {showMessage("Sprite: Attempt to write text to unlocked sprite");return;}
	static BYTE* font=NULL;
	static int charNo;

	BYTE* pScrn = pSurface+x+y*pitch;
	for(unsigned int cc=0;cc<strlen(s);cc++)
	{
		charNo=((int)(s[cc]));

		if (charNo==32) // space
		{
			for(int yy=0; yy<16; yy++)
				memset(pScrn+yy*pitch,'\0',16);
			pScrn+=16;
		}
		else
		{
			switch(charNo)
			{
			case '$' : font=Sprite::pFont+128+26*256; break;
			case '%' : font=Sprite::pFont+128+27*256; break;
			case '^' : font=Sprite::pFont+128+28*256; break;
			case '&' : font=Sprite::pFont+128+29*256; break;
			case '.' : font=Sprite::pFont+128+40*256; break;
			default :
				if (charNo>=120)
					font=Sprite::pFont+128+(charNo-79)*256;
				else if (charNo>=65 && charNo<=90)
					font=Sprite::pFont+128+(charNo-65)*256;
				else if (charNo>=48 && charNo<=57)
					font=Sprite::pFont+128+(charNo-18)*256;
			}

			if (font==NULL) return;

			for(int yy=0;yy<16;yy++)
			{
				memcpy(pScrn,font,16);
				pScrn+=pitch; font+=16;
			}
			pScrn+=16-pitch*16;
		}
	}
}

BOOL Sprite::drawText(int n,int x,int y)
{
	// As prevoius function but converts 'n' from integer to string first

	if (!bLocked) { showMessage("Sprite: Attempt to print to unlocked sprite"); return FALSE; }

	if (n<0) return FALSE;

	static char buffer[100];
	buffer[99]='\0';
	int ptr=99;

	if (n==0)
	{
		ptr--;
		buffer[ptr]=48;
	}
	else
		while(n!=0)
		{
			ptr--;
			buffer[ptr]=(char)(n%10+48);
			n=(int)(n/10);
		}

	drawText(buffer+ptr,x,y);
	return TRUE;
}

BOOL Sprite::drawText(float s, int x, int y)
{
	// Display floating point number to 2 decimal places accuracy

	if (!bLocked) { showMessage("Sprite: Attempt to print to unlocked sprite"); return FALSE; }

	static char buffer[100];

	buffer[99]='\0';
	int ptr=96;
	buffer[98]=(char)((((int)((s+0.005)*100)) %10)+48);
	buffer[97]=(char)((((int)((s+0.005)*10)) %10)+48);
	buffer[96]='.';

	int n=(int)(s+0.005);

	if (n==0)
	{
		ptr--;
		buffer[ptr]=48;
	}
	else
		while(n!=0)
		{
			ptr--;
			buffer[ptr]=(char)((n%10)+48);
			n=(int)(n/10);
		}

	drawText(buffer+ptr,x,y);
	return TRUE;
}

BOOL Sprite::drawText(char* s, int n, int x, int y)
{
	// Plots string 's' followed by integer 'n'

	drawText(s,x,y);
	drawText(n,x+strlen(s)*16,y);

	return TRUE;
}

BOOL Sprite::drawLine(int x1, int y1, int x2, int y2, int col)
{
	// Line drawing function
	// Only handles horizontal and vertical lines

	if (!bLocked) { showMessage("Sprite: Attempt to draw line to unlocked sprite"); return FALSE;}

	int bpp; // bytes per pixel
	DDPIXELFORMAT ddPixelFormat;
	ZeroMemory(&ddPixelFormat, sizeof(DDPIXELFORMAT));
	ddPixelFormat.dwSize = sizeof(DDPIXELFORMAT);
	pdds->GetPixelFormat(&ddPixelFormat);
	if (ddPixelFormat.dwRGBBitCount%8!=0 || ddPixelFormat.dwRGBBitCount>32) return FALSE;
	bpp = ddPixelFormat.dwRGBBitCount/8;

	if (y1==y2) // horizontal line
	{
		if (x1==x2) // point
			*(pSurface+y1*pitch+x1*bpp)=(BYTE)(col);
		else
		{
			if (x1<x2)
				memset(pSurface+y1*pitch+x1*bpp,col,(x2-x1)*bpp);
			else
				memset(pSurface+y1*pitch+x2*bpp,col,(x1-x2)*bpp);
		}
	}
	else if (x1==x2) // vertical line
		for(;y1!=y2; y1+=(y1>y2 ? -1 : 1))
		{
			*(pSurface+y1*pitch+x1*bpp)=(BYTE)(col);
			if (bpp>1) *(pSurface+y1*pitch+x1*bpp+1)=(BYTE)(col);
			if (bpp>2) *(pSurface+y1*pitch+x1*bpp+2)=(BYTE)(col);
			if (bpp>3) *(pSurface+y1*pitch+x1*bpp+3)=(BYTE)(col);
		}

	return TRUE;
}

BOOL Sprite::drawRect(int left, int top, int right, int bottom, int col)
{
	// Outline rectangles

	if (!bLocked) { showMessage("Attempt to draw line to unlocked sprite"); return FALSE;}

	drawLine(left,top,right,top,col);
	drawLine(right,top,right,bottom,col);
	drawLine(left,bottom,right,bottom,col);
	drawLine(left,top,left,bottom,col);

	return TRUE;
}

BOOL Sprite::drawFillRect(int x1, int y1, int x2, int y2, int col)
{
	// Solid colour rectangles

	if (!bLocked) { showMessage("Sprite: Attempt to write to unlocked sprite"); return FALSE; }

	if (y1==y2) // zero height? just draw line
	{
		drawLine(x1,y1,x2,y2,col);
		return TRUE;
	}

	if (y1>y2) // y co-ords wrong way round?
	{   int temp=y1; y1=y2; y2=temp;	}

	if (x1>x2) // x co-ords wrong way round?
	{	int temp=x1; x1=x2; x2=temp;	}

	for(;y1<y2; y1++)
		memset(pSurface+y1*pitch+x1,col,x2-x1);

	return TRUE;
}

void Sprite::setMask()
{	mask=TRUE;	}

void Sprite::clearMask()
{	mask=FALSE;	}

BOOL Sprite::restoreAll()
{
	if ( (lpDDSPrimary->Restore()) != DD_OK)
	{
		oneMessage("primary restore failed");
		return FALSE;
	}

	list.reset();
	Sprite* item=list.getNext();

	while(item!=NULL)
	{
		if (!(item->restore())) return FALSE;
		item=list.getNext();
	}

	Sprite::surfacesLost=TRUE;

	return TRUE;
}

BOOL Sprite::restore()
{
	if ( (pdds->Restore()) != DD_OK ) return FALSE;

    if (Sprite::noRestore) return TRUE;

	if (filename!=NULL)
	{
		if (DDReLoadBitmap(pdds,filename) != DD_OK)
		{
			showMessage("Error restoring sprite ",filename);
			return FALSE;
		}
	}
	else
		blacken();

	if (flipped) flip();

	return TRUE;
}

void Sprite::flip()
{
	int bpp; // bytes per pixel
	DDPIXELFORMAT ddPixelFormat;
	ZeroMemory(&ddPixelFormat, sizeof(DDPIXELFORMAT));
	ddPixelFormat.dwSize = sizeof(DDPIXELFORMAT);
	pdds->GetPixelFormat(&ddPixelFormat);
	if (ddPixelFormat.dwRGBBitCount%8!=0 || ddPixelFormat.dwRGBBitCount>32) return;
	bpp = ddPixelFormat.dwRGBBitCount/8;

	BYTE temp[4];
	BYTE* ptr=lock();
	for(int y=0; y<height; y++)
		for(int x=0; x<width/2; x++)
		{
			memcpy(temp, ptr+y*pitch+x*bpp, bpp);
			memcpy(ptr+y*pitch+x*bpp, ptr+y*pitch+(width-1-x)*bpp, bpp);
			memcpy(ptr+y*pitch+(width-1-x)*bpp, temp, bpp);
//			temp=*(ptr+y*pitch+x);
//			*(ptr+y*pitch+x) = *(ptr+y*pitch+width-1-x);
//			*(ptr+y*pitch+width-1-x) = temp;
		}
	unlock();
	flipped=!flipped;
}

BOOL Sprite::createSurfaces()
{
	HRESULT ddrval;
	// Get exclusive mode
	if (Sprite::fullScreen==TRUE)
	{
		ddrval = Sprite::lpDD->SetCooperativeLevel( Sprite::hwnd,
			DDSCL_EXCLUSIVE | DDSCL_FULLSCREEN );
		if( ddrval != DD_OK )
		{
			showMessage("Sprite: Could not get exclusive mode");
			return FALSE;
		}
		// Set the video mode to scrnWidth*scrnHeight*8 bits per pixel
		ddrval = (Sprite::lpDD)->SetDisplayMode( scrnWidth, scrnHeight, Sprite::hicolour?16:8);
		if( ddrval != DD_OK )
			{ showMessage("SetDisplayMode failed");return FALSE;	}
		// Create the primary surface with 1 back buffer
		DDSURFACEDESC ddsd;
		ddsd.dwSize = sizeof( ddsd );
		ddsd.dwFlags = DDSD_CAPS | DDSD_BACKBUFFERCOUNT;
		ddsd.ddsCaps.dwCaps = DDSCAPS_PRIMARYSURFACE |
							  DDSCAPS_FLIP |
							  DDSCAPS_COMPLEX;
		ddsd.dwBackBufferCount = 1;
		if (Sprite::lpDDSPrimary!=NULL)
		{
			Sprite::lpDDSPrimary->Release();
			Sprite::lpDDSPrimary=NULL;
		}
		if (Sprite::lpDDSBack!=NULL)
		{
			Sprite::lpDDSBack->Release();
			Sprite::lpDDSBack=NULL;
		}
		ddrval = Sprite::lpDD->CreateSurface( &ddsd, &Sprite::lpDDSPrimary, NULL );
		if( ddrval != DD_OK )
		{	showMessage("Sprite: Could not create primary surface"); return FALSE;}
		DDSCAPS ddscaps;
		ddscaps.dwCaps = DDSCAPS_BACKBUFFER;
		ddrval = Sprite::lpDDSPrimary->GetAttachedSurface(&ddscaps, &Sprite::lpDDSBack);
		if( ddrval != DD_OK )
		{	showMessage("Sprite: Could not retrieve secondary surface"); return FALSE;}
		Sprite::lpDDClipper=NULL;
	}
	else
	{
		ddrval = Sprite::lpDD->SetCooperativeLevel( hwnd, DDSCL_NORMAL);
		if( ddrval != DD_OK )
		{
			showMessage("Sprite: Could not set cooperative level");
			return FALSE;
		}
		// Create the primary surface
		DDSURFACEDESC ddsd;
		ZeroMemory(&ddsd,sizeof(DDSURFACEDESC));
		ddsd.dwSize = sizeof( DDSURFACEDESC );
		ddsd.dwFlags = DDSD_CAPS;
		ddsd.ddsCaps.dwCaps = DDSCAPS_PRIMARYSURFACE;

		if (Sprite::lpDDSPrimary!=NULL)
		{
			Sprite::lpDDSPrimary->Release();
			Sprite::lpDDSPrimary=NULL;
		}
		ddrval = Sprite::lpDD->CreateSurface( &ddsd, &Sprite::lpDDSPrimary, NULL );
		if( ddrval != DD_OK )
		{
			showMessage("Sprite: Could not create primary surface");
			return FALSE;
		}

		// now make an offscreen surface to assemble images onto
		if (Sprite::lpDDSBack!=NULL)
		{	Sprite::lpDDSBack->Release(); Sprite::lpDDSBack=NULL; }

		ZeroMemory(&ddsd, sizeof(ddsd));
		ddsd.dwSize = sizeof(ddsd);
		ddsd.dwFlags = DDSD_CAPS | DDSD_HEIGHT |DDSD_WIDTH;
		ddsd.ddsCaps.dwCaps = DDSCAPS_OFFSCREENPLAIN;
		ddsd.dwWidth = Sprite::scrnWidth; ddsd.dwHeight = Sprite::scrnHeight;

		if (Sprite::lpDD->CreateSurface(&ddsd, &Sprite::lpDDSBack, NULL) != DD_OK)
			showMessage("Sprite::CreateSurfaces: Error creating DD surface");

		ddrval = Sprite::lpDD->CreateClipper( 0, &Sprite::lpDDClipper, NULL );
		if( ddrval != DD_OK )
		{	showMessage("Sprite: Could not create Clipper"); return FALSE; }

		ddrval = Sprite::lpDDClipper->SetHWnd( 0, Sprite::hwnd );
		if( ddrval != DD_OK )
		{	showMessage("Sprite: Could not set Clipper hwnd"); return FALSE; }

		ddrval = Sprite::lpDDSPrimary->SetClipper( Sprite::lpDDClipper );
		if( ddrval != DD_OK )
		{	showMessage("Sprite: Could not lock Clipper"); return FALSE; }
	}

	// clear front and back screens
	DDBLTFX ddbltfx;
	ddbltfx.dwSize = sizeof(ddbltfx);
	ddbltfx.dwFillColor =0;
	if (Sprite::lpDDSPrimary->Blt(	NULL, NULL, NULL, DDBLT_COLORFILL, &ddbltfx)!=DD_OK)
		showMessage("error flood filling front surface");

	if (Sprite::lpDDSBack->Blt(	NULL, NULL, NULL,  DDBLT_COLORFILL, &ddbltfx)!=DD_OK)
		showMessage("error flood filling back surface");
	return TRUE;
}

BOOL Sprite::setSize(BOOL fs, int xs, int ys)
{
//showMessage("starting sprite.setsize...");
		Sprite::fullScreen=fs;
		Sprite::scrnWidth=xs; Sprite::scrnHeight=ys;
		createSurfaces();
//showMessage("Created surfaces");
		restoreAll();
//showMessage("Sprite.setsize done ok");
		return TRUE;
}

BOOL Sprite::loseSurfaces()
{
	if (!Sprite::fullScreen)
	{
		if (Sprite::lpDDSBack!=NULL)
		{
			Sprite::lpDDSBack->Release();
			Sprite::lpDDSBack=NULL;
		}
	}
	else
	{
		Sprite::lpDDSBack=NULL;
	}

	if (Sprite::lpDDSPrimary!=NULL)
	{
		Sprite::lpDDSPrimary->Release();
		Sprite::lpDDSPrimary=NULL;
	}
			return TRUE;
}

void Sprite::cls()
{
	Sprite* b = new Sprite;
	b->setSize(Sprite::scrnWidth,Sprite::scrnHeight);
	b->draw(0,0);
	if (Sprite::fullScreen)
	{
		flipScreens();
		b->draw(0,0);
	}
	delete b;
}

void Sprite::colourize(SPRITECOL col, Sprite* colours, int xx, int yy, int magnify)
{

	static unsigned int bpp=0; // bytes per pixel
	static BYTE cols[5][4]; // [noOfColours][max size of pixel]

	DDPIXELFORMAT ddPixelFormat;
	ZeroMemory(&ddPixelFormat, sizeof(DDPIXELFORMAT));
	ddPixelFormat.dwSize = sizeof(DDPIXELFORMAT);
	pdds->GetPixelFormat(&ddPixelFormat);
	if (ddPixelFormat.dwRGBBitCount%8!=0 || ddPixelFormat.dwRGBBitCount>32) return;

	if (bpp != ddPixelFormat.dwRGBBitCount/8) // recalculate colours
	{
		bpp = ddPixelFormat.dwRGBBitCount/8;
		BYTE* ptr=colours->lock();
		for(int c=0; c<5; c++)
			memcpy(cols[c],ptr+(int)(yy*colours->getPitch()+xx*bpp+c*bpp*magnify),bpp);
		colours->unlock();
	}

	BYTE* ptr=lock();
	for(int x=0; x<width; x++)
		for(int y=0; y<height; y++)
		{
			if ( *(ptr+y*pitch+x*bpp)>0 )
				*(ptr+y*pitch+x*bpp)=cols[col][0];
			if ( bpp>1 && *(ptr+y*pitch+x*bpp+1)>0 )
				*(ptr+y*pitch+x*bpp+1)=cols[col][1];
			if ( bpp>2 && *(ptr+y*pitch+x*bpp+2)>0 )
				*(ptr+y*pitch+x*bpp+2)=cols[col][2];
			if ( bpp>3 && *(ptr+y*pitch+x*bpp+3)>0 )
				*(ptr+y*pitch+x*bpp+3)=cols[col][3];
		}

	unlock();
}

void Sprite::report()
{
	list.reset();
	Sprite* item=list.getNext();
	char mess[512];
	ZeroMemory(mess,512);
	DDSURFACEDESC surfacedesc;

	while(item!=NULL)
	{
		ZeroMemory(&surfacedesc,sizeof(surfacedesc));
		surfacedesc.dwSize=sizeof(surfacedesc);
		item->pdds->GetSurfaceDesc(&surfacedesc);
		mess[0]='\0';
		if (item->filename!=NULL) strcat(mess,item->filename);
		else strcat(mess,"No name");
		strcat(mess,"\n");
		if (surfacedesc.ddsCaps.dwCaps&DDSCAPS_SYSTEMMEMORY)
			strcat(mess,"System memory\n");
		else
			strcat(mess,"Video memory\n");
		showMessage(mess);
		item=list.getNext();
	}
}

void Sprite::showMode()
{
/*	DDSURFACEDESC surfacedesc;
	ZeroMemory(&surfacedesc,sizeof(surfacedesc));
	surfacedesc.dwSize=sizeof(surfacedesc);
	surfacedesc.dwFlags=DDSD_REFRESHRATE;
	Sprite::lpDDSPrimary->GetSurfaceDesc(&surfacedesc);*/
	unsigned long refreshRate=0;
	HRESULT res=Sprite::lpDD->GetMonitorFrequency(&refreshRate);
	switch(res)
	{
	case DD_OK : showMessage("Refresh rate (hz) = ",refreshRate/100); break;
	case DDERR_UNSUPPORTED: showMessage("Cannot determine refresh rate");break;
	default: showMessage("GetMonitorFrequency failed");
	}
}

